# coding: utf-8
from inspect import isclass
from .helper import NO_THING
from . helper.base import HelperBase
from . rules import IRule
from .exceptions import CancelEventError, ReservedAttributeError, RuleError
from typing import Iterable, List, Optional, Callable
from collections import UserList
from typing import Any, Dict, List, Optional, Set, Callable, Union
VERSION = __version__ = '2.3.0'

# region class HelperArgs


class HelperArgs(HelperBase):
    """
    Helper class that provides KwArgs arguments
    """

    def __init__(self, key: str, ** kwargs):
        """
        Constructor

        Args:
            key (str): Key Arg

        Keyword Arguments:
            default (obj, optional): Default arg. Default ``NO_THING``
            field (str, optional): field arg. Default ``None``
            require (bool, optional): require arg. Default ``False``
            rules_all (Iterable, optional): rules_any list. Default Empty List.
            rules_any (Iterable, optional): rules_all list. Default Empty List.
            types (set, optional): types arg. Default Empty set
        """
        self._key: str = ''
        self._field = None
        self._require = False
        self._types = set()
        self._default = NO_THING
        self._rules_all = []
        self._rules_any = []
        self.key = key
        keys = ('field', 'require', 'types',
                'default', 'rules_all', 'rules_any')
        for key in keys:
            if key in kwargs:
                setattr(self, key, kwargs[key])

    def to_dict(self) -> dict:
        '''Gets a dictionary representation of current instance fields'''
        arg = {'key': self.key, 'require': self.require}
        if self.field is not None:
            arg['field'] = self.field
        if self.types is not None and len(self.types) > 0:
            arg['types'] = [itm for itm in self.types]
        if self.default is not NO_THING:
            arg['default'] = self.default
        if self.rules_all is not None and len(self.rules_all) > 0:
            arg['rules_all'] = self.rules_all
        if self.rules_any is not None and len(self.rules_any) > 0:
            arg['rules_any'] = self.rules_any
        return arg

    # region Properties

    @property
    def default(self) -> object:
        """
        Default Value

        :getter: Gets default value
        :setter: Sets default value
        """
        return self._default

    @default.setter
    def default(self, value: object) -> None:
        self._default = value
        return None

    @property
    def key(self) -> str:
        """
        Key Value

        :getter: Gets Key Value
        :setter: Sets key value
        """
        return self._key

    @key.setter
    def key(self, value: str) -> None:
        self._is_prop_str(value=value, prop_name='key', raise_error=True)
        self._key = value.strip()
        return None

    @property
    def field(self) -> Union[str, None]:
        """
        Field Value

        :getter: Gets field value
        :setter: Sets field value
        """
        return self._field

    @field.setter
    def field(self, value: Union[str, None]) -> None:
        if value is None:
            self._field = None
            return None
        self._is_prop_str(value=value, prop_name='field', raise_error=True)
        self._field = value.strip()
        return None

    @property
    def require(self) -> bool:
        """
        Require Value

        :getter: Gets require value
        :setter: Sets require value
        """
        return self._require

    @require.setter
    def require(self, value: bool) -> None:
        self._is_prop_bool(value=value, prop_name='require', raise_error=True)
        self._require = value
        return None

    @property
    def types(self) -> Set[type]:
        """
        Types values

        :getter: Gets types values
        :setter: Sets types values
        """
        return self._types

    @types.setter
    def types(self, value: Union[list, set, tuple]) -> None:
        if not isinstance(value, (list, set, tuple)):
            self._prop_error('types', value, 'Iterable')
        if isinstance(value, set):
            self._types = value
            return None
        self._types = set([itm for itm in value])
        return None

    @property
    def rules_all(self) -> List[Callable[[IRule], bool]]:
        """
        Rules values

        :getter: Gets rules_all
        :setter: Sets rules_all
        """
        return self._rules_all

    @rules_all.setter
    def rules_all(self, value: List[Callable[[IRule], bool]]) -> None:
        self._isinstance_prop(value=value, prop_name='rules_all',
                              prop_type=list, raise_error=True)
        self._rules_all = value

    @property
    def rules_any(self) -> List[Callable[[IRule], bool]]:
        """
        Rules values

        :getter: Gets rules_any
        :setter: Sets rules_any
        """
        return self._rules_any

    @rules_any.setter
    def rules_any(self, value: List[Callable[[IRule], bool]]) -> None:
        self._isinstance_prop(value=value, prop_name='rules_any',
                              prop_type=list, raise_error=True)
        self._rules_any = value
    # endregion Properties
# endregion class HelperArgs

# region class AssignBuilder


class AssignBuilder(UserList):
    '''Helper class for building list to use with "KwargsHelper.Assing() method'''

    def __init__(self) -> None:
        """
        Constructor
        """
        super().__init__(initlist=None)
        self._keys = set()

    def append(self, key: str, field: Optional[str] = None, require: bool = False, default: Optional[object] = None, types: Optional[List[type]] = None, rules_all: Optional[List[Callable[[IRule], bool]]] = None, rules_any: Optional[List[Callable[[IRule], bool]]] = None):
        """
        Appends dictionary item of parameters to list

        Args:
            key (str): the key of the key, value pair.
            field (Optional[str], optional): the name of the field.. Defaults to ``None``.
            require (bool, optional): Determins if ``key`` is required. Defaults to ``False``.
            default (Optional[object], optional): default value to assign if key value is missing. Defaults to ``None``.
            types (Optional[List[type]], optional): list of one or more types that the value of the key value pair must match. Defaults to ``None``.
            rules (Optional[List[Callable[[IRule], bool]]], optional): Rules to apply. Defaults to ``None``.

        Raises:
            TypeError: if ``key`` is not instance of``str``.
            ValueError: if ``key`` is empty or whitespace str.
            ValueError: if ``key`` has already exist.

        """

        if not isinstance(key, str):
            raise TypeError(self._get_type_error_method_msg(
                method_name='append', arg=key, arg_name='key', expected_type='str'
            ))
        _key = key.strip()
        if len(_key) == 0:
            raise ValueError(self._get_value_error_msg(
                method_name='append', arg=key, arg_name='key',
                msg="can not be empty or whitespace"
            ))
        if _key in self._keys:
            raise ValueError(
                self._get_value_error_msg(
                    method_name='append', arg=key, arg_name='key',
                    msg='already exist.'
                )
            )
        _args = HelperArgs(key=_key)
        _args.require = require
        if field is not None:
            _args.field = field
        if default is not None:
            _args.default = default
        if types is not None:
            _args.types = set(types)
        if rules_all is not None:
            _args.rules_all = rules_all
        if rules_any is not None:
            _args.rules_any = rules_any

        super().append(_args)
        self._keys.add(_key)
        return None

    def append_helper(self, helper: HelperArgs):
        """
        Appends dictionary item of parameters to list

        Args:
            helper (HelperArgs): parameters to append

        Raises:
            TypeError: if ``helper`` is not instance of ``HelperArgs``
            ValueError: if ``helper.key`` is empty string.
            ValueError: if ``helper.key`` already exist.
        """
        if not isinstance(helper, HelperArgs):
            raise TypeError(self._get_type_error_method_msg(
                method_name='append_helper', arg=helper, arg_name='helper', expected_type='HelperArgs'
            ))
        if len(helper.key) == 0:
            raise ValueError(self._get_value_error_msg(
                method_name='append_helper', arg=helper, arg_name='helper.key',
                msg="can not be empty or whitespace"
            ))

        if helper.key in self._keys:
            raise ValueError(
                self._get_value_error_msg(
                    method_name='append_helper', arg=helper, arg_name='key',
                    msg='already exist.'
                )
            )
        super().append(helper)
        self._keys.add(helper.key)

    def remove(self, item: HelperArgs) -> None:
        """
        Removes an instance of ``HelperArgs`` from this instance

        Args:
            item (HelperArgs): Object to remove

        Raises:
            TypeError: if ``item`` is not instance of ``HelperArgs``

        Returns:
            [obj]: ``None``
        """
        if item is None:
            return None
        if not isinstance(item, HelperArgs):
            raise TypeError(self._get_type_error_method_msg(
                method_name='remove', arg=item, arg_name='item', expected_type='HelperArgs'
            ))
        _key = item.key
        super().remove(item)
        self._keys.remove(_key)
        return None

    def extend(self, other: 'AssignBuilder') -> None:
        """
        Extents this instance by merging values from other instance of ``AssignBuilder``.

        Args:
            other (AssignBuilder): instance to merge

        Raises:
            NotImplementedError: if ``other`` is not instance of ``AssignBuilder``.
        """
        if not isinstance(other, AssignBuilder):
            raise NotImplementedError(
                f"{self.__class__.__name__}.extend() only supports extending by instances of 'AssignBuilder'")
        for item in other:
            key: str = item.key
            if not key in self._keys:
                super().append(item)
                self._keys.add(key)

    def _get_type_error_method_msg(self, method_name: str, arg: object, arg_name: str, expected_type: str) -> str:
        result = f"{self.__class__.__name__}.{method_name}() arg '{arg_name}' is expecting type of '{expected_type}'. Got type of '{type(arg).__name__}'"
        return result

    def _get_value_error_msg(self, method_name: str, arg: object, arg_name: str, msg: str) -> str:
        result = f"{self.__class__.__name__}.{method_name}() arg '{arg_name}' {msg}"
        return result

    # region dunder methods
    def __getitem__(self, i: int) -> HelperArgs:
        return super().__getitem__(i)

    def __setitem__(self, i: int, helper: HelperArgs):
        if not isinstance(helper, HelperArgs):
            raise TypeError(self._get_type_error_method_msg(
                method_name='__setitem__', arg=helper, arg_name='helper', expected_type='HelperArgs'
            ))
        if i < 0 or i >= len(self):
            raise IndexError
        current = self[i]
        if helper.key != current.key:
            if helper.key in self._keys:
                raise ValueError(
                    self._get_value_error_msg(
                        method_name='__setitem__', arg=helper, arg_name='key',
                        msg='already exist.'
                    ))
        super().__setitem__(i, helper)
    # endregion dunder methods
# endregion class AssignBuilder

# region Event Args


class BeforeAssignAutoEventArgs:
    """
    Before Assign Auto Event Args

    See Also:
        :doc:`../usage/KwargsHelper/auto_assign_callback`    
    """

    def __init__(self, key: str, value: object, field: str, originator: object):
        """
        Constructor

        Args:
            key (str): Key arg
            value (object): Value to be assigned
            field (str): Field that ``value`` is to be assigned to.
            originator (object): Object that ``value`` is to be assigned to.
            all_rules (bool): Determines if all rules or any rules are to be matched
        """
        self._key = key
        self._field_value = value
        self._cancel = False
        self._originator = originator
        self._field_name = field

    # region Properties
    @property
    def key(self) -> str:
        '''
        Gets the key of the key value pair
        '''
        return self._key

    @property
    def field_name(self) -> str:
        '''
        The field that will be assigned representing ``key``

        :getter: Gets the field that will be assigned representing ``key``
        :setter: Setts the field that will be assigned representing ``key``
        '''
        return self._field_name

    @field_name.setter
    def field_name(self, value: str) -> str:
        self._field_name = str(value).strip()

    @property
    def field_value(self) -> object:
        '''
        The value to be blindly assigned to property

        :getter: Gets the value to be blindly assigned to property
        :setter: Sets the value to be blindly assigned to property
        '''
        return self._field_value

    @field_value.setter
    def field_value(self, value: object) -> None:
        self._field_value = value

    @property
    def originator(self) -> object:
        '''Gets object that attributes assigned/modified for '''
        return self._originator

    @property
    def cancel(self) -> bool:
        """
        Cancel arg. If ``True`` process will be canceled.

        :getter: Gets cancel arg
        :setter: Sets cancel arg
        """
        return self._cancel

    @cancel.setter
    def cancel(self, value: bool) -> None:
        self._cancel = bool(value)
    # endregion Properties


class AfterAssignAutoEventArgs:
    """
    After Assign Auto Event Args

    See Also:
        :doc:`../usage/KwargsHelper/auto_assign_callback` 
    """

    def __init__(self, key: str, originator: object):
        """
        Constructor

        Args:
            key (str): key arg
            originator (object): object that originated event
        """
        self._key = key
        self._field_value: object = None
        self._originator = originator
        self._field_name: str = ''
        self._success = False
        self._canceled = False

    # region Properties
    @property
    def key(self) -> str:
        '''
        Gets the key of the key value pair
        '''
        return self._key

    @property
    def field_name(self) -> str:
        '''
        The field that will be assigned representing ``key``.

        :getter: Gets the field that will be assigned representing ``key``.
        :setter: Sets the field that will be assigned representing ``key``.
        '''
        return self._field_name

    @property
    def field_value(self) -> object:
        return self._field_value

    @property
    def originator(self) -> object:
        '''Gets object that attributes assigned/modified for '''
        return self._originator

    @property
    def success(self) -> bool:
        '''Get assigning of attribue/value succeeded'''
        return self._success

    @property
    def canceled(self) -> bool:
        '''Get if assigment was canceled by before events'''
        return self._canceled
    # endregion Properties


class BeforeAssignEventArgs:
    """
    Before assign event args

    See Also:
        :doc:`../usage/KwargsHelper/callback` 
    """

    def __init__(self, help_args: HelperArgs, originator: object):
        """
        Constructor

        Args:
            help_args (HelperArgs): HelperArgs object
            originator (object): Origanating object
        """
        self._helper_args = help_args
        self._originator = originator
        self._field_name: str = ''
        self._field_value: object = None
        self._cancel = False
    # region Properties

    @property
    def field_name(self) -> str:
        """
        The name of the field that value will be assigned

        :getter: Gets the name of the field that value will be assigned.
        :setter: Sets the name of the field that value will be assigned.
        """
        return self._field_name

    @field_name.setter
    def field_name(self, value: str) -> None:
        self._field_name = str(value)

    @property
    def field_value(self) -> object:
        """
        The value that will be assigned ``field_name``.

        :getter: Gets the value that will be assigned ``field_name``.
        :setter: Sets the value that will be assigned ``field_name``.
        """
        return self._field_value

    @field_value.setter
    def field_value(self, value: object) -> None:
        self._field_value = value

    @property
    def key(self) -> str:
        '''Gets the key currently being read'''
        return self._helper_args.key

    @property
    def helper_args(self) -> HelperArgs:
        ''''Get the args used to for modify/creating attribute'''
        return self._helper_args

    @property
    def originator(self) -> object:
        '''Gets object that attributes assigned/modified for '''
        return self._originator

    @property
    def cancel(self) -> bool:
        """
        Cancel arg. If ``True`` process will be canceled.

        :getter: Gets cancel arg
        :setter: Sets cancel arg
        """
        return self._cancel

    @cancel.setter
    def cancel(self, value: bool) -> None:
        self._cancel = bool(value)
    # endregion Properties


class AfterAssignEventArgs:
    """
    After Assign Event Args

    See Also:
        :doc:`../usage/KwargsHelper/callback` 
    """

    def __init__(self, help_args: HelperArgs, originator: object) -> None:
        """
        Constructor

        Args:
            help_args (HelperArgs): HelperArgs object
            originator (object): Origanating object
        """
        self._helper_args = help_args
        self._field_name: str = ''
        self._field_value: object = None
        self._originator = originator
        self._rules_passed = True
        self._canceled = False
        self._success = False

    # region Properties

    @property
    def key(self) -> str:
        '''Gets the key currently being read'''
        return self._helper_args.key

    @property
    def field_name(self) -> str:
        '''The name of the field that value was assigned'''
        return self._field_name

    @property
    def field_value(self) -> object:
        '''The value that is assigned to `field_name`'''
        return self._field_value

    @property
    def helper_args(self) -> HelperArgs:
        ''''Get the args used to for modify/creating attribute'''
        return self._helper_args

    @property
    def originator(self) -> object:
        '''Gets object that attributes assigned/modified for '''
        return self._originator

    @property
    def rules_passed(self) -> bool:
        '''Get if all applied rules passed'''
        return self._rules_passed

    @property
    def canceled(self) -> bool:
        '''Get if assigment was canceled by before events'''
        return self._canceled

    @property
    def success(self) -> bool:
        """
        Determinins if assigning of attribue/value succeeded

        :getter: Get assigning of attribue/value succeeded
        :setter: Sets assigning of attribue/value succeeded
        """
        return self._success

    @success.setter
    def success(self, value: bool) -> None:
        self._success = bool(value)
    # endregion Properties
# endregion Event Args

# region class KwargsHelper


class KwargsHelper(HelperBase):
    """
    kwargs helper class. Assigns attributes to class with various checks

    Example:
        .. include:: ../inc/ex/KwargsHelper_basic_cls.rst

        .. include:: ../inc/ex/KwargsHelper_basic_ex.rst
    """
    # region init

    def __init__(self, originator: object, obj_kwargs: dict, **kwargs):
        """
        Constructor

        Args:
            originator (object): object that attributes are assigned to via ``assign()`` method. This is usually a class.
            obj_kwargs (dict): The dictionary of key value args used to set values of attributes.
                Often passed in as ``obj_kwargs = {**kwargs}``

        Keyword Arguments:
            field_prefix (str, optional): sets the ``field_prefix`` property. Default ``_``.
            name (str, optional): sets the `field_prefix` property.
                Default is the name of ``originator`` object.
            cancel_error (bool, optional): sets the `cancel_error` property. Default ``True``.
            rule_error (bool, optional): sets the ``rule_error`` property. Default ``True``.
            assign_true_not_required (bool, optional): sets the ``assign_true_not_required`` property.
                Default ``True``.
            type_instance_check (bool, optional): If ``True`` and :py:meth:`.KwargsHelper.assign`` arg ``types`` is set
                then values will be tested also for isinstance rather then just type check if type check is ``False``.
                If ``False`` then values willl only be tested as type.
                Default ``True``

        Raises:
            TypeError: if any arg is not of the correct type.
        """
        if not isinstance(obj_kwargs, dict):
            raise TypeError(self._get_type_error_method_msg(
                method_name='__init__', arg=obj_kwargs,
                arg_name='obj_kwargs', expected_type=dict
            ))
        self._obj: object = originator
        self._callbacks = None
        self._kwargs = obj_kwargs
        self._auto_assigned = False
        self._keys: set = set(self._kwargs.keys())
        m_name = '__init__'
        key = 'field_prefix'
        if key in kwargs:
            self._field_prefix = kwargs[key]
            self._is_arg_str(
                method_name=m_name, arg=self._field_prefix, arg_name=key, raise_error=True)
        else:
            self._field_prefix = '_'
        key = 'name'
        if key in kwargs:
            self._name = kwargs[key]
            self._is_arg_str(
                method_name=m_name, arg=self._name, arg_name=key, raise_error=True)
        else:
            self._name = type(originator).__name__

        key = 'rule_error'
        if key in kwargs:
            self._rule_error: bool = kwargs[key]
            self._is_arg_bool(
                method_name=m_name, arg=self._rule_error, arg_name=key, raise_error=True)
        else:
            self._rule_error: bool = True

        key = 'rule_test_before_assign'
        if key in kwargs:
            self._rule_test_early: bool = kwargs[key]
            self._is_arg_bool(
                method_name=m_name, arg=self._rule_test_early, arg_name=key, raise_error=True)
        else:
            self._rule_test_early = True

        key = 'cancel_error'
        if key in kwargs:
            self._cancel_error: bool = kwargs[key]
            self._is_arg_bool(
                method_name=m_name, arg=self._cancel_error, arg_name=key, raise_error=True)
        else:
            self._cancel_error: bool = True

        key = 'assign_true_not_required'
        if key in kwargs:
            self._assign_true_not_required: bool = kwargs[key]
            self._is_arg_bool(
                method_name=m_name, arg=self._assign_true_not_required, arg_name=key, raise_error=True)
        else:
            self._assign_true_not_required: bool = True

        key = 'type_instance_check'
        if key in kwargs:
            self._type_instance_check: bool = kwargs[key]
            self._is_arg_bool(
                method_name=m_name, arg=self._type_instance_check, arg_name=key, raise_error=True)
        else:
            self._type_instance_check: bool = True

    # endregion init

    # region Public Methods
    def auto_assign(self, types: Optional[Iterable[type]] = None, rules_all: Optional[Iterable[Callable[[IRule], bool]]] = None, rules_any: Optional[Iterable[Callable[[IRule], bool]]] = None) -> bool:
        """
        Assigns all of the key, value pairs of ``obj_kwargs`` passed into constructor to ``originator``,
        unless the event is canceled in :py:class:`.BeforeAssignAutoEventArgs` then key,
        value pair will be added automacally to ``originator``.
        
        Args:
            types (Iterable[type], optional): a type list of one or more types that the value of the key value pair must match.
                For example if all values are required to be only ``str`` then ``types=[str]``.
                If all values are required to be ``str`` or ``int`` then ``types=[str, int]``.

                If ``types`` is omitted then values can be any type unless there is a rule in ``rules`` that is otherwise.
                Defaults to ``None``.

                See also: :doc:`../usage/KwargsHelper/assign_type`
            rules_all (Iterable[Callable[[IRule], bool]], optional): List of rules that must be passed before assignment can take place.
                If ``types`` is included then ``types`` takes priority over this arg.
                All rules must validate as ``True`` before assignment takes place.
                Defaults to ``None``.

                See also: :doc:`../usage/KwargsHelper/assign_rules`
            rules_any (Iterable[Callable[[IRule], bool]], optional): List of rules that must be passed before assignment can take place.
                If ``types`` is included then ``types`` takes priority over this arg.
                Any rule that validates as ``True`` results in assignment taking place.
                Defaults to ``None``.

                See also: :doc:`../usage/KwargsHelper/assign_rules`

        Returns:
            bool: ``True`` of all key, value pairs are added; Otherwise, ``False``.

        Raises:
            RuleError: If :py:attr:`~KwargsHelper.rule_error` is ``True`` and Validation of rules fails.
            TypeError: If validation of ``types`` fails.

        Note:
            Call back events are supported via :py:meth:`~.KwargsHelper.add_handler_before_assign_auto`
            and :py:meth:`~.KwargsHelper.add_handler_after_assign_auto` methods.

        See Also:
            * :doc:`../usage/KwargsHelper/auto_assign`
            * :doc:`../usage/KwargsHelper/auto_assign_callback`
            * :py:meth:`~.KwargsHelper.assign`
        """
        valid = True
        if self._rule_test_early is True:
            valid = self._auto_assign_validation(types=types,
                                                rules_all=rules_all,
                                                rules_any=rules_any)
            if valid is False:
               return False
        if self._is_auto_assign_handlers():
            valid = self._auto_assign_with_cb()
        else:
            valid = self._auto_assign_no_cb()
        if self._rule_test_early is False:
            valid = valid & self._auto_assign_validation(types=types,
                                                 rules_all=rules_all,
                                                 rules_any=rules_any)
        return valid

    def assign(self, key: str, field: Optional[str] = None, require: bool = False, default: Optional[object] = NO_THING, types: Optional[Iterable[type]] = None, rules_all: Optional[Iterable[Callable[[IRule], bool]]] = None, rules_any: Optional[Iterable[Callable[[IRule], bool]]] = None) -> bool:
        """
        Assigns attribute value to ``obj`` passed in to constructor. Attributes are created if they do not exist.

        Args:
            key (str): the key of the key, value pair that is required or optional in ``obj_kwargs`` passed into to constructor.
            field (str, optional): the name of the field to assign a value. If ``field``
                is omitted then field name is built using ``instance.field_prefix`` + ``key``.
                If included then ``instance.field_prefix`` will be ignored.
                Defaults to ``None``.

                See also: :doc:`../usage/KwargsHelper/assign_field`
            require (bool, optional): Determins if ``key`` is required to be in ``obj_kwargs`` passed into to constructor.
                if ``default`` is passed in then ``require`` is ignored.
                Defaults to ``False``.

                See also: :doc:`../usage/KwargsHelper/assign_require`
            default (object, optional): default value to assign to key attribute if no value is found in
                ``obj_kwargs`` passed into to constructor.
                If ``default`` is passed in then ``require`` is ignored.
                Defaults to ``NO_THING`` which will result in default being ignored.

                See also: :doc:`../usage/KwargsHelper/assign_default`
            types (Iterable[type], optional): a type list of one or more types that the value of the key value pair must match.
                For example if a value is required to be only ``str`` then ``types=[str]``.
                If value is required to be ``str`` or ``int`` then ``types=[str, int]``.
                In this example if value is not type ``str`` then ``TypeError`` is raised.
                If ``types`` is omitted then a value can be any type unless there is a rule in ``rules`` that is otherwise.
                Defaults to ``None``.

                See also: :doc:`../usage/KwargsHelper/assign_type`
            rules_all (Iterable[Callable[[IRule], bool]], optional): List of rules that must be passed before assignment can take place.
                If ``types`` is included then ``types`` takes priority over this arg.
                All rules must validate as ``True`` before assignment takes place.
                Defaults to ``None``.

                See also: :doc:`../usage/KwargsHelper/assign_rules`
            rules_any (Iterable[Callable[[IRule], bool]], optional): List of rules that must be passed before assignment can take place.
                If ``types`` is included then ``types`` takes priority over this arg.
                Any rule that validates as ``True`` results in assignment taking place.
                Defaults to ``None``.

                See also: :doc:`../usage/KwargsHelper/assign_rules`

        Returns:
            bool: ``True`` if attribute assignment is successful; Otherwise, ``False``

        Raises:
            RuleError: If :py:attr:`~KwargsHelper.rule_error` is ``True`` and Validation of
                ``rules_all`` or ``rules_any`` fails.
            TypeError: If validation of ``types`` fails.

        See Also:
            * :doc:`../usage/KwargsHelper/assign_field`
            * :doc:`../usage/KwargsHelper/assign_default`
            * :doc:`../usage/KwargsHelper/assign_type`
            * :doc:`../usage/KwargsHelper/assign_rules`
            * :py:meth:`~.KwargsHelper.auto_assign`
        """
        m_name = 'assign'
        self._is_arg_str_empty_null(
            method_name=m_name, arg=key, arg_name='key', raise_error=True)
        self._is_arg_bool(
            method_name=m_name, arg=require, arg_name='require', raise_error=True)

        if types == None:
            types = []
        if rules_all == None:
            rules_all = []
        if rules_any == None:
            rules_any = []
        _args = HelperArgs(key=key, require=require, rules_all=rules_all, rules_any=rules_any)
        _args.field = field
        # _args.require = require
        _args.types = set(types)
        if default is not NO_THING:
            _args.default = default
        before_args = BeforeAssignEventArgs(_args, self._obj)

        after_args = AfterAssignEventArgs(_args, self._obj)
        self._assign(args=_args, before_args=before_args,
                     after_args=after_args)
        self._on_after_assign(after_args)
        result = after_args.success

        if result == False and self._assign_true_not_required == True and require == False:
            result = True
        return result

    def assign_helper(self, helper: HelperArgs) -> bool:
        """
        Assigns attribute value using instance of HelperArgs. See :py:meth:`~.KwargsHelper.assign` method.

        Args:
            helper (HelperArgs): instance to assign.

        Returns:
            bool: ``True`` if attribute assignment is successful; Otherwise, ``False``.
        """
        self._isinstance_method(method_name='assign_helper', arg=helper,
                                arg_name='helper', arg_type=HelperArgs, raise_error=True)
        d = helper.to_dict()
        return self.assign(**d)

    def is_key_existing(self, key: str) -> bool:
        """
        Gets if the key exist in  kwargs dictionary passed in to the constructor by ``obj_kwargs`` arg.

        Args:
            key (str): Key value to check for existance.

        Returns:
            bool: ``True`` if ``key`` exist; Otherwise, ``False``.
        """
        valid_key = self._is_arg_str_empty_null(
            method_name='is_key_existing', arg=key, arg_name='key', raise_error=False)
        if valid_key == False:
            return False
        return key in self._kwargs
    # endregion Public Methods

    # region private methods

    # region internal assign methods
    def _auto_assign_no_cb(self) -> bool:
        for k, v in self._kwargs.items():
            field_name = f"{self._field_prefix}{k}"
            setattr(self._obj, field_name, v)
            # self._obj.__dict__[field_name] = v
            self._remove_key(key=k)
        return True

    def _auto_assign_with_cb(self) -> bool:
        '''
        Assigns all key, value pairs blindly unless interupted by event cancel
        @return: `True` if all key, values have been added; Otherwise, `False`
        '''
        self._auto_assigned == True
        result = True
        for k, v in self._kwargs.items():
            field_name = f"{self._field_prefix}{k}"
            before_args = BeforeAssignAutoEventArgs(
                key=k, value=v, field=field_name, originator=self._obj)
            self._on_before_assign_auto(event_args=before_args)
            _field = before_args.field_name
            _value = before_args.field_value
            after_args = AfterAssignAutoEventArgs(key=k, originator=self._obj)
            after_args._canceled = before_args.cancel
            if before_args.cancel == True:
                if self._cancel_error == True:
                    raise CancelEventError(
                        f"{self.__class__.__name__}.auto_assign() canceled in 'BeforeAssignBlindEventArgs'")
                else:
                    result = False
            else:
                # self._obj.__dict__[_field] = _value
                setattr(self._obj, _field, _value)
                after_args._success = True
                after_args._field_name = _field
                after_args._field_value = _value
                self._remove_key(key=k)
            self._on_after_assign_auto(event_args=after_args)
        return result

    def _assign(self, args: HelperArgs, before_args: BeforeAssignEventArgs, after_args: AfterAssignEventArgs) -> None:
        result = False
        key = args.key
        if key in self._kwargs:
            value = self._kwargs[key]
            self._validate_type(key=key, value=value, types=args.types)
            if args.field:
                result = self._setattr(
                    args.field, value, before_args, after_args, args=args)
            else:
                result = self._setattr(f"{self._field_prefix}{key}",
                                       value, before_args, after_args, args=args)
        else:
            if args.default is not NO_THING:
                if args.field:
                    result = self._setattr(args.field, args.default,
                                           before_args, after_args, args=args)
                else:
                    result = self._setattr(f"{self._field_prefix}{key}",
                                           args.default, before_args, after_args, args=args)
            elif args.require:
                # only test for required when default is not included
                raise ValueError(f"{self._name} arg '{key}' is required")
        after_args._success = result
        after_args._success
        return None

    # endregion internal assign methods

    def _get_formated_types(self, types: Iterable[type]) -> str:
        result = ''
        for i, t in enumerate(types):
            if i > 0:
                result = result + ' | '
            result = f"{result}{t}"
        return result

    def _setattr(self, field: str, value: object, before_args: BeforeAssignEventArgs, after_args: AfterAssignEventArgs, args: HelperArgs) -> bool:
        before_args.field_name = field
        before_args.field_value = value

        self._on_before_assign(event_args=before_args)

        after_args._canceled = before_args.cancel
        if before_args.cancel == True:
            if self._cancel_error == False:
                return False
            raise CancelEventError(
                f"{self.__class__.__name__}.assign() canceled in 'BeforeAssignEventArgs'")
        _field = before_args.field_name
        _value = before_args.field_value
        result = True
        if self._rule_test_early:
            result = self._validate_assign_rules(
                args=args, field=_field, value=_value, after_args=after_args)
            if result == False:
                return result
        setattr(self._obj, _field, _value)
        # self._obj.__dict__[_field] = _value
        self._remove_key(args.key)
        if self._rule_test_early == False:
            result = self._validate_assign_rules(
                args=args, field=_field, value=_value, after_args=after_args)
            if result == False:
                return result
        after_args._field_name = _field
        after_args._field_value = _value
        return result

    # region internal validation methods
    def _auto_assign_validation(self, types: Optional[Iterable[type]] = None, rules_all: Optional[Iterable[Callable[[IRule], bool]]] = None, rules_any: Optional[Iterable[Callable[[IRule], bool]]] = None) -> bool:
        if isinstance(types, Iterable):
            self._validate_types(types)
        valid = True
        if isinstance(rules_all, Iterable):
            valid = self._validate_auto_assign_rules_all(
                rules=rules_all)
        if valid is False:
            return False
        if isinstance(rules_any, Iterable):
            valid = self._validate_auto_assign_rules_any(
                rules=rules_any)
        return valid

    def _validate_type(self, key: str, value: object, types: Iterable[type]):
        def _is_type_instance(_types: Set[type], _value):
            result = False
            for t in _types:
                if isinstance(_value, t):
                    result = True
                    break
            return result
        if len(types) == 0:
            return
        if not type(value) in types:
            # object such as PosixPath inherit from more than on class (Path, PurePosixPath)
            # testing if PosixPath is type of Path is False.
            # for this reason will do an instace check as well. isinstance(_posx, Path) is True
            is_valid_type = False
            if self._type_instance_check == True and _is_type_instance(types, value):
                is_valid_type = True

            if is_valid_type is False:
                msg = f"{self._name} arg '{key}' is expected to be of '{self._get_formated_types(types)}' but got '{type(value).__name__}'"
                raise TypeError(msg)

    def _validate_types(self, types: Iterable[type]):
        if len(types) == 0:
            return
        for k, v in self._kwargs.items():
           self._validate_type(key=k, value=v, types=types)

    def _validate_assign_rules(self, args: HelperArgs, field: str, value: object,  after_args: AfterAssignEventArgs) -> bool:
        result = self._validate_assign_rules_all(
            args=args, field=field, value=value, after_args=after_args)
        if result is False:
            return False
        result = self._validate_assign_rules_any(
            args=args, field=field, value=value, after_args=after_args)
        return result

    def _validate_auto_assign_rules_all(self, rules: Iterable[IRule]) -> bool:
        result = True
        if len(rules) > 0:
            for k, v in self._kwargs.items():
                result = self._validate_rules_all(rules=rules,
                                                  key=k,
                                                  field=self._field_prefix + k,
                                                  value=v)
                if result is False:
                    break
        return result

    def _validate_auto_assign_rules_any(self, rules: Iterable[IRule]) -> bool:
        result = True
        if len(rules) > 0:
            for k, v in self._kwargs.items():
                rules_valid = self._validate_rules_any(rules=rules,
                                                  key=k,
                                                  field=self._field_prefix + k,
                                                  value=v)
                if rules_valid is False:
                    result = False
                    break
        return result

    def _validate_assign_rules_all(self, args: HelperArgs, field: str, value: object,  after_args: AfterAssignEventArgs) -> bool:
        # if all rules pass then validations is considered a success
        result = self._validate_rules_all(rules=args.rules_all, key=after_args.key, field=field, value=value)
        after_args._rules_passed = result
        return result

    def _validate_assign_rules_any(self, args: HelperArgs, field: str, value: object,  after_args: AfterAssignEventArgs) -> bool:
        # if any rule passes then validations is considered a success
        result = self._validate_rules_any(rules=args.rules_any, key=after_args.key, field=field, value=value)
        after_args._rules_passed = result
        return result

    def _validate_rules_all(self, rules: Iterable[IRule], key: str, field: str, value: object) -> bool:
        # if all rules pass then validations is considered a success
        result = True
        if len(rules) > 0:
            for rule in rules:
                if not isclass(rule) or not issubclass(rule, IRule):
                    raise TypeError('Rules must implement IRule')
                rule_instance: IRule = rule(
                    key=key, name=field, value=value, raise_errors=self._rule_error, originator=self._obj)
                try:
                    result = result & rule_instance.validate()
                except Exception as e:
                    raise RuleError(
                        err_rule=rule, rules_all=rules, arg_name=key, errors=e) from e
                if result is False:
                    break
        return result

    def _validate_rules_any(self, rules: Iterable[IRule], key: str, field: str, value: object) -> bool:
        # if any rule passes then validations is considered a success
        error_lst = []
        failed_rules = []
        result = True
        if len(rules) > 0:
            for rule in rules:
                if not isclass(rule) or not issubclass(rule, IRule):
                    raise TypeError('Rules must implement IRule')
                rule_instance: IRule = rule(
                    key=key, name=field, value=value, raise_errors=self._rule_error, originator=self._obj)
                rule_valid = False
                try:
                    rule_valid = rule_instance.validate()
                except Exception as e:
                    error_lst.append(e)
                    failed_rules.append(rule)
                    rule_valid = False
                result = rule_valid
                if rule_valid is True:
                    break
        if result is False and self._rule_error is True and len(error_lst) > 0:
            # raise the first error in error list
            raise RuleError(rules_any=rules,
                            err_rule=failed_rules[0], arg_name=key, errors=error_lst) from error_lst[0]
        return result

    # endregion internal validation methods

    def _remove_key(self, key: str) -> bool:
        '''
        Removes a key from the internal keys.
        @return: `True` if key was removed; Otherwise, `False`
        '''
        if key in self._keys:
            self._keys.remove(key)
            return True
        return False

    def _is_auto_assign_handlers(self) -> bool:
        '''
        Gets if any handler are set for Auto Assign
        '''
        if self._callbacks is None:
            return False
        if 'on_before_assign_auto' in self._callbacks:
            return True
        if 'on_after_assign_auto' in self._callbacks:
            return True
        return False
    # endregion private methods

    # region callback funcs

    def _on(self, event_name, callback):
        if self._callbacks is None:
            self._callbacks = {}

        if event_name not in self._callbacks:
            self._callbacks[event_name] = [callback]
        else:
            self._callbacks[event_name].append(callback)

    def _trigger(self, event_name, eventArgs):
        if self._callbacks is not None and event_name in self._callbacks:
            for callback in self._callbacks[event_name]:
                callback(self, eventArgs)

    # endregion callback funcs

    # region raise events
    def _on_before_assign(self, event_args: BeforeAssignEventArgs):
        self._trigger("on_before_assign", event_args)

    def _on_after_assign(self, event_args: AfterAssignEventArgs):
        self._trigger("on_after_assign", event_args)

    def _on_before_assign_auto(self, event_args: BeforeAssignAutoEventArgs):
        self._trigger("on_before_assign_auto", event_args)

    def _on_after_assign_auto(self, event_args: AfterAssignAutoEventArgs):
        self._trigger("on_after_assign_auto", event_args)
    # endregion raise events

     # region Handlers
    def add_handler_before_assign(self, callback: Callable[['KwargsHelper', BeforeAssignEventArgs], None]):
        """
        Add handler before assign

        Args:
            callback (Callable[['KwargsHelper', BeforeAssignEventArgs], None]): Callback Method

        See Also:
            :doc:`../usage/KwargsHelper/callback`
        """
        self._on("on_before_assign", callback)

    def add_handler_after_assign(self, callback: Callable[['KwargsHelper', AfterAssignEventArgs], None]):
        """
        Add handler afer assign

        Args:
            callback (Callable[['KwargsHelper', AfterAssignEventArgs], None]): Callback Method

        See Also:
            :doc:`../usage/KwargsHelper/callback`
        """
        self._on("on_after_assign", callback)

    def add_handler_before_assign_auto(self, callback: Callable[['KwargsHelper', BeforeAssignAutoEventArgs], None]):
        """
        Adds handler for before assigne auto

        Args:
            callback (Callable[['KwargsHelper', BeforeAssignAutoEventArgs], None]): Callback Method

        See Also:
            :doc:`../usage/KwargsHelper/callback`
        """
        self._on("on_before_assign_auto", callback)

    def add_handler_after_assign_auto(self, callback: Callable[['KwargsHelper', AfterAssignAutoEventArgs], None]):
        """
        Adds handler for after assign auto

        Args:
            callback (Callable[['KwargsHelper', AfterAssignAutoEventArgs], None]): Callback Method

        See Also:
            :doc:`../usage/KwargsHelper/callback`
        """
        self._on("on_after_assign_auto", callback)
    # endregion

    # region Properties
    @property
    def assign_true_not_required(self) -> bool:
        """
        Determines ``assign_true_not_required`` Option.
        If ``True`` then and a non-required arg is assigned via :py:meth:`~.KwargsHelper.assign`.
        then :py:meth:`~.KwargsHelper.assign` returns ``True`` even if the arg failed to be applied.
        In an ``after callback`` method set by :py:meth:`~.KwargsHelper.add_handler_after_assign`
        success in :py:attr:`.AfterAssignEventArgs.success` property is ``False`` if arg was not assigned.
        Default ``True``

        :getter: Gets option value.
        :setter: Sets option value.
        """
        return self._assign_true_not_required

    @assign_true_not_required.setter
    def assign_true_not_required(self, value: bool) -> None:
        self._is_prop_bool(
            value=value, prop_name='assign_true_not_required', raise_error=True)
        self._assign_true_not_required = value
        return None

    @property
    def cancel_error(self) -> bool:
        """
        Determins if an error will be raised if cancel is set in
        :py:class:`BeforeAssignEventArgs` of a callback. Default ``True``.

        :getter: Gets cancel_error option.
        :setter: Sets cancel_error option.
        """
        return self._cancel_error

    @cancel_error.setter
    def cancel_error(self, value: bool) -> None:
        self._is_prop_bool(
            value=value, prop_name='cancel_error', raise_error=True)
        self._cancel_error = value
        return None

    @property
    def name(self) -> str:
        """
        Determins name option.
        ``name`` that represents the ``originator`` in error messages. Default: ``type(originator)__name__``

        :getter: Gets name option.
        :setter: Sets name option
        """
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._is_prop_str(
            value=value, prop_name='name', raise_error=True)
        self._name = value
        return None

    @property
    def field_prefix(self) -> str:
        """
        Determines the field prefix option.
        The prefix use when setting attributes: Default: ``_``.
        To add args without a prefix set the value ``field_prefix = ''``
        This parameter is ignored when ``field`` is used in :py:meth:`~.KwargsHelper.assign` method such as: ``assign(msg='hi', field='message')``

        :getter: Gets the field prefix option.
        :setter: Sets the field prefix option.
        """
        return self._field_prefix

    @field_prefix.setter
    def field_prefix(self, value: str) -> None:
        self._is_prop_str(
            value=value, prop_name='field_prefix', raise_error=True)
        self._field_prefix = value
        return None

    @property
    def originator(self) -> object:
        '''
        Gets originator option

        object that attributes are assigned to via :py:meth:`~.KwargsHelper.assign` method.
        This is usually a class.
        '''
        return self._obj

    @property
    def rule_error(self) -> bool:
        """
        Determins rule_error option.
        Default ``True``

        :getter: Gets rule_error option.
        :setter: Sets rule_error option.
        """
        return self._rule_error

    @rule_error.setter
    def rule_error(self, value: bool) -> None:
        self._is_prop_bool(
            value=value, prop_name='rule_error', raise_error=True)
        self._rule_error = value
        return None

    @property
    def rule_test_before_assign(self) -> bool:
        """
        Determines rule_test_before_assign option.
        If ``True`` rule testing will occur before assign value to attribute.
        If ``True`` and :py:attr:`~.KwargsHelper.rule_error` is ``True`` then rule errors will prevent assigning value.
        If ``False`` then attribute values will be assigned even if rules does not validate.
        Validation can still fail and errors can still be raised, except now the validation
        will take place after attribute value has been assigned.
        Default: ``True``

        :getter: Gets rule_test_before_assign option.
        :setter: Sets rule_test_before_assign option.
        """
        return self._rule_test_early

    @rule_test_before_assign.setter
    def rule_test_before_assign(self, value: bool) -> None:
        self._is_prop_bool(
            value=value, prop_name='rule_test_before_assign', raise_error=True)
        self._rule_test_early = value
        return None

    @property
    def kw_args(self) -> Dict[str, Any]:
        '''Gets the kwargs dictionary passed in to the constructor by ``obj_kwargs`` arg'''
        return self._kwargs

    @property
    def unused_keys(self) -> Set[str]:
        '''
        Gets any unused keys passed into constructor via ``obj_kwargs``

        This would be a set of keys that were never used passed into the constructor.
        '''
        return self._keys
    # endregion Properties

# endregion class KwargsHelper

# region class KwArg


class KwArg:
    '''Class for assigning kwargs to autogen fields with type checking and testing'''
    class _KwArgInternal(HelperBase):
        def __init__(self, orig, **kwargs):
            self.helper_instance = KwargsHelper(
                originator=orig, obj_kwargs={**kwargs})
            self.helper_instance.field_prefix = ''

    _RESERVED_INTERNAL_FIELDS: Set[str] = set(
        ['_kw_arg_internal', '_KwArgInternal', 'kwargs_helper',
         '__init__', 'is_attribute_exist', 'is_key_existing',
         'kw_unused_keys', 'kw_assign'])

    def __init__(self, **kwargs):
        """
        Constructor

        Keyword Arguments:
            kwargs (dict): dictionary of args

        Example:
            .. include:: ../inc/ex/KwArg_basic.rst
        """
        self._kw_arg_internal = KwArg._KwArgInternal(orig=self, **kwargs)

    def kw_assign(self, key: str, field: Optional[str] = None, require: bool = False, default: Optional[object] = NO_THING, types: Optional[List[type]] = None, rules_all: Optional[List[Callable[[IRule], bool]]] = None, rules_any: Optional[List[Callable[[IRule], bool]]] = None) -> bool:
        """
        Assigns attribute value to current instance passed in to constructor. Attributes automatically.

        Args:
            key (str): the key of the key, value pair that is required or optional in ``kwargs`` passed into to constructor.
            all_rules (bool, optional): Determines if all rules or any rules are to be matched.
                If ``True`` then all rules included in ``rules`` must be valid to be considered a success.
                If ``False`` then any rule included in ``rules`` that is valid is considered a success.
                Default ``False``.
            field (str, optional): the name of the field to assign a value. 
                if ``field`` is omitted then field name is built using ``key``.
                If included then ``kwargs_helper.field_prefix`` will be ignored.
                Defaults to **Empty string**.
                
                See also: :doc:`../usage/KwArg/kw_assign_field`,
                :py:attr:`.KwArg.kwargs_helper`
            require (bool, optional): Determins if ``key`` is required to be in `kwargs` passed into to constructor.
                if ``default`` is passed in then ``require`` is ignored.
                Defaults to ``False``.
                
                See also: :doc:`../usage/KwArg/kw_assign_require`
            default (object, optional): default value to assign to key attribute if no value is
                found in ``kwargs`` passed into to constructor.
                If ``default`` is passed in then ``require`` is ignored.
                Defaults to ``NO_THING``.
                
                See also: :doc:`../usage/KwArg/kw_assign_default`
            types (List[type], optional): a type list of one or more types that the value of the
                key value pair must match.
                For example if a value is required to be only ``str`` then ``types=[str]``.
                In this example if value is not type ``str`` then ``TypeError`` is raised
                If value is required to be `str` or `int` then ``types=[str, int]``.
                Defaults to ``None``.
                
                See also: :doc:`../usage/KwArg/kw_assign_type`
            rules_all (List[Callable[[IRule], bool]], optional): List of rules that must be passed before assignment can take place.
                If ``types`` is included then ``types`` takes priority over this arg.
                All rules must validate as ``True`` before assignment takes place.
                Defaults to ``None``.
                
                See also: :doc:`../usage/KwArg/kw_assign_rules`
            rules_any (List[Callable[[IRule], bool]], optional): List of rules that must be passed before assignment can take place.
                If ``types`` is included then ``types`` takes priority over this arg.
                Any rule that validates as ``True`` results in assignment taking place.
                Defaults to ``None``.
                
                See also: :doc:`../usage/KwArg/kw_assign_rules`

        Raises:
            RuleError: If ``kwargs_helper.rule_error`` is ``True`` and Validation of
                ``rules_all`` or ``rules_any`` fails.
            TypeError: If validation of ``types`` fails.
            ReservedAttributeError: if ``key`` is a reserved keyword
            ReservedAttributeError: if ``field`` is a reserved keyword

        Returns:
            bool: ``True`` if attribute assignment is successful; Otherwise, ``False``

        See Also:
            * :doc:`../usage/KwArg/kw_assign_default`
            * :doc:`../usage/KwArg/kw_assign_field`
            * :doc:`../usage/KwArg/kw_assign_require`
            * :doc:`../usage/KwArg/kw_assign_rules` 
            * :doc:`../usage/KwArg/kw_assign_type`
        """
        m_name = 'assign'
        self._kw_arg_internal._is_arg_str_empty_null(
            method_name=m_name, arg=key, arg_name='key', raise_error=True)
        if field is None:
            if key in KwArg._RESERVED_INTERNAL_FIELDS:
                raise ReservedAttributeError(
                    f"{self.__class__.__name__}.{key} is a reserved keyword. Try using the 'field' arg for a different field name")
        else:
            if field in KwArg._RESERVED_INTERNAL_FIELDS:
                raise ReservedAttributeError(
                    f"{self.__class__.__name__}.{field} is a reserved keyword. Try using a differne field name.")
        return self._kw_arg_internal.helper_instance.assign(key=key, field=field, require=require, default=default, types=types, rules_all=rules_all, rules_any=rules_any)

    # region Public Methods
    def is_attribute_exist(self, attrib_name: str) -> bool:
        """
        Gets if ``attrib_name`` exist the current instance.

        Use this method when:

        * When assigning a key that is not required it may not exist in the current instance.
        * When assing key that are required then the field will exist in the current instance.

        Args:
            attrib_name (str): This is usually the ``key`` value of :py:meth:`~.KwArg.kw_assign`
                or ``field`` value of :py:meth:`~.KwArg.kw_assign`.

        Returns:
            bool: ``True`` if ``attrib_name`` exist in current instance; Otherwise, ``False``.
        """
        # if an key is assigned and not required it may not exist in curren instance
        valid_key = self._kw_arg_internal._is_arg_str_empty_null(
            method_name='has_attribute', arg=attrib_name, arg_name='key', raise_error=False)
        if valid_key == False:
            return False
        _key = attrib_name.strip()
        return hasattr(self, _key)

    def is_key_existing(self, key: str) -> bool:
        '''
        Gets if the key exist in current kwargs.
        Basically shortcut for `key in kwargs`
        '''
        return self._kw_arg_internal.helper_instance.is_key_existing(key)
    # endregion Public Methods

    # region Properties

    @property
    def kwargs_helper(self) -> KwargsHelper:
        '''Get instance of KwargsHelper used to add fields current instance'''
        return self._kw_arg_internal.helper_instance

    @property
    def kw_unused_keys(self) -> Set[str]:
        '''
        Gets any unused keys passed into constructor via ``**kwargs``

        This would be a set of keys that were never used passed into the constructor.
        '''
        return self._kw_arg_internal.helper_instance.unused_keys
    # endregion Properties

# endregion class KwArg



# coding: utf-8
from kwhelp.rules import RuleAttrExist
from typing import Optional
from abc import ABC, abstractmethod
# region class HelperBase

class HelperBase(ABC):
    """Helper Base class"""
    @abstractmethod
    def __init__(self):
        '''Class Constructor'''

    # region private methods

    def _get_type_error_method_msg(self, method_name: str, arg: object, arg_name: str, expected_type: str) -> str:
        result = f"{self.__class__.__name__}.{method_name}() arg '{arg_name}' is expecting type of '{expected_type}'. Got type of '{type(arg).__name__}'"
        return result

    def _get_value_error_msg(self, method_name: str, arg: object, arg_name: str, msg: str) -> str:
        result = f"{self.__class__.__name__}.{method_name}() arg '{arg_name}' {msg}"
        return result

    # region Property Helpers
    def _get_type_error_prop_msg(self, prop_name: str, value: object, expected_type: str) -> str:
        result = f"{self.__class__.__name__}.{prop_name} is expecting type of '{expected_type}'. Got type of '{type(value).__name__}'"
        return result

    def _isinstance_prop(self, value: object, prop_name: str, prop_type: object, raise_error: Optional[bool] = False):
        result = isinstance(value, prop_type)
        if result == False and raise_error == True:
            self._prop_error(prop_name=prop_name,
                             value=value, expected_type=self._get_name_type_obj(prop_type))
        return result

    def _prop_error(self, prop_name: str, value: object, expected_type: str):
        raise TypeError(self._get_type_error_prop_msg(
            prop_name=prop_name, value=value, expected_type=expected_type
        ))

    def _is_prop_str(self, value: object, prop_name: str, raise_error: Optional[bool] = False) -> bool:
        result = self._isinstance_prop(
            value=value, prop_name=prop_name, prop_type=str, raise_error=raise_error)
        return result

    def _is_prop_bool(self, value: object, prop_name: str, raise_error: Optional[bool] = False) -> bool:
        result = self._isinstance_prop(
            value=value, prop_name=prop_name, prop_type=bool, raise_error=raise_error)
        return result

    def _is_prop_int(self, value: object, prop_name: str, raise_error: Optional[bool] = False) -> bool:
        result = self._isinstance_prop(
            value=value, prop_name=prop_name, prop_type=int, raise_error=raise_error)
        return result
    # endregion Property Helpers

    # region method Arg Helpers
    def _isinstance_method(self, method_name: str, arg: object, arg_name: str, arg_type: object, raise_error: Optional[bool] = False) -> bool:
        result = isinstance(arg, arg_type)
        if result == False and raise_error == True:
            self._arg_type_error(self._get_type_error_method_msg(
                method_name=method_name, arg=arg, arg_name=arg_name,
                expected_type=self._get_name_type_obj(
                    arg_type)
            ))
        return result

    def _is_arg_str(self, method_name: str, arg: object, arg_name: str, raise_error: Optional[bool] = False) -> bool:
        result = self._isinstance_method(
            method_name=method_name, arg=arg, arg_name=arg_name, arg_type=str, raise_error=raise_error)
        return result
    
    def _is_arg_str_empty_null(self, method_name: str, arg: object, arg_name: str, raise_error: Optional[bool] = False) -> bool:
        result = self._is_arg_str(
            method_name=method_name, arg=arg, arg_name=arg_name, raise_error=raise_error)
        if result == False:
            return result
        _arg = str(arg).strip()
        if len(_arg) == 0:
            if raise_error:
                raise ValueError(self._get_value_error_msg(
                    method_name=method_name, arg=arg, arg_name=arg_name,
                    msg='empty or whitespace string is not allowed'
                ))
            result = False
        return result

    def _is_arg_bool(self, method_name: str, arg: object, arg_name: str, raise_error: Optional[bool] = False) -> bool:
        result = self._isinstance_method(
            method_name=method_name, arg=arg, arg_name=arg_name, arg_type=bool, raise_error=raise_error)
        return result
    # endregion method Arg Helpers

    def _arg_type_error(self, method_name: str, arg: object, arg_name: str, expected_type: str):
        raise TypeError(self._get_type_error_method_msg(
            method_name=method_name, arg=arg, arg_name=arg_name, expected_type=expected_type
        ))

    def _get_name_type_obj(self, obj: object) -> str:
        '''
        Gets the name of an object instance name or type name
        '''
        if isinstance(obj, type):
            return str(obj.__name__)
        return str(obj.__class__.__name__)
    # endregion private methods
# endregion class HelperBase



# coding: utf-8
from inspect import isclass
from typing import Iterable, Optional, Union
from ..rules import IRule
from ..helper import is_iterable
from ..exceptions import RuleError
class TypeChecker:
    """Class that validates args match a give type"""

    def __init__(self, *args: type, **kwargs): 
        """
        Constructor

        Other Arguments:
            args (type): One or more types used for Validation purposes.

        Keyword Arguments:
            raise_error: (bool, optional): If ``True`` then an error will be raised if a :py:meth:`~.TypeChecker.validate` fails:
                Othwewise :py:meth:`~.TypeChecker.validate` will return a boolean value indicationg success or failure.
            type_instance_check (bool, optional): If ``True`` then :py:meth:`~.TypeChecker.validate` args
                are tested also for isinstance if type does not match, rather then just type check if type is a match.
                If ``False`` then values willl only be tested as type.
                Default ``True``
        """
        self._types = [arg for arg in args]

        key = 'raise_error'
        if key in kwargs:
            self._raise_error: bool = bool(kwargs[key])
        else:
            self._raise_error: bool = True

        key = 'type_instance_check'
        if key in kwargs:
            self._type_instance_check: bool = bool(kwargs[key])
        else:
            self._type_instance_check: bool = True

    def _get_formated_types(self) -> str:
        result = ''
        for i, t in enumerate(self._types):
            if i > 0:
                result = result + ' | '
            result = f"{result}{t}"
        return result

    def _validate_type(self, value: object,  key: Union[str, None] = None):
        def _is_type_instance(_types: Iterable[type], _value):
            result = False
            for t in _types:
                if isinstance(_value, t):
                    result = True
                    break
            return result
        result = True
        if not type(value) in self._types:
            # object such as PosixPath inherit from more than on class (Path, PurePosixPath)
            # testing if PosixPath is type of Path is False.
            # for this reason will do an instace check as well. isinstance(_posx, Path) is True
            is_valid_type = False
            if self._type_instance_check == True and _is_type_instance(self._types, value):
                is_valid_type = True

            if is_valid_type is True:
                result = True
            else:
                result = False
                if self._raise_error is True:
                    if key is None:
                        msg = f"Arg Value is expected to be of '{self._get_formated_types()}' but got '{type(value).__name__}'"
                    else:
                        msg = f"Arg '{key}' is expected to be of '{self._get_formated_types()}' but got '{type(value).__name__}'"
                    raise TypeError(msg)
        return result

    def validate(self, *args, **kwargs) -> bool:
        """
        Validates all ``*args`` and all ``**kwargs`` against ``types`` that are passed
        into constructor.

        Returns:
            bool: ``True`` if all ``*args`` and all ``**kwarg`` match a type; Otherwise;
            ``False``.

        Raises:
            TypeError: if ``raise_error`` is ``True`` and validation fails.
        """
        if len(self._types) == 0:
            return True
        result = True
        for arg in args:
            result = result & self._validate_type(value=arg)
            if result is False:
                break
        if result is False:
            return result
        for k, v in kwargs.items():
            result = result & self._validate_type(value=v, key=k)
            if result is False:
                break
        return result

    # region Properties
    @property
    def type_instance_check(self) -> bool:
        """
        Determines if instance checking is done with type checking.

        If ``True`` then :py:meth:`~.TypeChecker.validate`` args
        are tested also for isinstance if type does not match, rather then just type check if type is a match.
        If ``False`` then values willl only be tested as type.

        :getter: Gets type_instance_check value
        :setter: Sets type_instance_check value
        """
        return self._type_instance_check

    @type_instance_check.setter
    def type_instance_check(self, value: bool) -> bool:
        self._type_instance_check = bool(value)

    @property
    def raise_error(self) -> bool:
        """
        Determines if errors will be raised during validation

        If ``True`` then errors will be raised when validation fails.

        :getter: Gets if errors can be raised.
        :setter: Sets if errors can be raised.
        """
        return self._raise_error

    @raise_error.setter
    def raise_error(self, value: bool) -> bool:
        self._raise_error = bool(value)
    
    @property
    def types(self) -> Iterable[type]:
        """
        Gets the types passed into constructor that are used for validating args
        """
        return self._types
    # endregion Properties


class RuleChecker:
    """Class that validates args match a given rule"""

    def __init__(self, rules_all: Optional[Iterable[IRule]] = None, rules_any: Optional[Iterable[IRule]] = None, ** kwargs):
        """
        Constructor

        Args:
            rules_all (Iterable[IRule], optional): List of rules that must all be matched. Defaults to ``None``.
            rules_any (Iterable[IRule], optional): List of rules that any one must be matched. Defaults to ``None``.

        Keyword Arguments:
            raise_error (bool, optional): If ``True`` then rules can raise errors when validation fails.
                Default ``False``.

        Raises:
            TypeError: If ``rule_all`` is not an iterable object
            TypeError: If ``rule_any`` is not an iterable object
        """
        if rules_all is None:
            self._rules_all = []
            self._len_all = 0
        else:
            if is_iterable(rules_all) == False:
                raise TypeError(
                    "rules_all arg must be an iterable object such as list or tuple.")
            self._rules_all = rules_all
            self._len_all = len(self._rules_all)
        if rules_any is None:
            self._rules_any = []
            self._len_any = 0
        else:
            if is_iterable(rules_any) == False:
                raise TypeError(
                    "rules_aany arg must be an iterable object such as list or tuple.")
            self._rules_any = rules_any
            self._len_any = len(self._rules_any)
            
        
        key = 'raise_error'
        if key in kwargs:
            self._raise_error: bool = bool(kwargs[key])
        else:
            self._raise_error: bool = True

    # region internal validation methods

    def _validate_rules_all(self, key: str, field: str, value: object) -> bool:
        # if all rules pass then validations is considered a success
        valid_arg = self._is_valid_arg(key)
        if valid_arg:
            _key = key
            _field = field
        else:
            _key = 'arg'
            _field = 'arg'
           
        result = True
        if self._len_all > 0:
            for rule in self._rules_all:
                if not isclass(rule) or not issubclass(rule, IRule):
                    raise TypeError('Rules must implement IRule')
                rule_instance: IRule = rule(
                    key=_key, name=_field, value=value, raise_errors=self._raise_error, originator=self)
                try:
                    result = result & rule_instance.validate()
                except Exception as e:
                    arg_name = _key if valid_arg else None
                    raise RuleError(
                        err_rule=rule, rules_all=self._rules_all, arg_name=arg_name, errors=e) from e
                if result is False:
                    break
        return result

    def _validate_rules_any(self, key: str, field: str, value: object) -> bool:
        # if any rule passes then validations is considered a success
        valid_arg = self._is_valid_arg(key)
        if valid_arg:
            _key = key
            _field = field
        else:
            _key = 'arg'
            _field = 'arg'
        error_lst = []
        result = True
        failed_rules = []
        if self._len_any > 0:
            for rule in self._rules_any:
                if not isclass(rule) or not issubclass(rule, IRule):
                    raise TypeError('Rules must implement IRule')
                rule_instance: IRule = rule(
                    key=_key, name=_field, value=value, raise_errors=self._raise_error, originator=self)
                rule_valid = False
                try:
                    rule_valid = rule_instance.validate()
                except Exception as e:
                    error_lst.append(e)
                    failed_rules.append(rule)
                    rule_valid = False
                result = rule_valid
                if rule_valid is True:
                    break
        if result is False and self._raise_error is True and len(error_lst) > 0:
            # raise the last error in error list
            arg_name = _key if valid_arg else None
            raise RuleError(rules_any=self._rules_any,
                            err_rule=failed_rules[0], arg_name=arg_name, errors=error_lst) from error_lst[0]
            
        return result

    def _is_valid_arg(self, arg:str) -> bool:
        return arg != '..arg'
    
    # endregion internal validation methods

    def validate_all(self, *args, **kwargs) -> bool:
        """
        Validates all. All ``*args`` and ``**kwargs`` must match :py:attr:`~.RuleChecker.rules_all`

        Returns:
            bool: ``True`` if all ``*args`` and ``**kwargs`` are valid; Otherwise, ``False``

        Raises:
            Exception: If :py:attr:`~.RuleChecker.raise_error` is ``True`` and validation Fails.
                The type of exception raised is dependend on the :py:class:`.IRule` that caused validation
                failure. Most rules raise a ``ValueError`` or a ``TypeError``.
        """
        if self._len_all == 0:
            return True
        result = True
        for arg in args:
            result = result & self._validate_rules_all(
                key="..arg", field="arg", value=arg)
            if result is False:
                break
        if result is False:
            return result
        for k, v in kwargs.items():
            result = result & self._validate_rules_all(key=k, field=k, value=v)
            if result is False:
                break
        return result

    def validate_any(self, *args, **kwargs) -> bool:
        """
        Validates any. All ``*args`` and ``**kwargs`` must match on ore more of :py:attr:`~.RuleChecker.rules_any`

        Returns:
            bool: ``True`` if all ``*args`` and ``**kwargs`` are valid; Otherwise, ``False``

        Raises:
            Exception: If :py:attr:`~.RuleChecker.raise_error` is ``True`` and validation Fails.
                The type of exception raised is dependend on the :py:class:`.IRule` that caused validation
                failure. Most rules raise a ``ValueError`` or a ``TypeError``.
        """
        if self._len_any == 0:
            return True
        result = True
        for arg in args:
            result = self._validate_rules_any(
                key="..arg", field="arg", value=arg)
            if result is False:
                break
        if result is False:
            return result
        for k, v in kwargs.items():
            result = result & self._validate_rules_any(key=k, field=k, value=v)
            if result is False:
                break
        return result

    # region Properties
    @property
    def rules_all(self) -> Iterable[IRule]:
        """
        Gets rules passed into ``rules_all`` of constructor used for validation of args.
        """
        return self._rules_all

    @property
    def rules_any(self) -> Iterable[IRule]:
        """
        Gets rules passed into ``rules_any`` of constructor used for validation of args.
        """
        return self._rules_any

    @property
    def raise_error(self) -> bool:
        """
        Determines if errors will be raised during validation

        If ``True`` then errors will be raised when validation fails.
        Default value is ``True``.

        :getter: Gets if errors can be raised.
        :setter: Sets if errors can be raised.
        """
        return self._raise_error

    @raise_error.setter
    def raise_error(self, value: bool) -> bool:
        self._raise_error = bool(value)
    # endregion Properties



# coding: utf-8
from collections.abc import Iterator

class Singleton(type):
    """Singleton abstrace class"""
    _instances = {}
    # https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(
                Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class NoThing(metaclass=Singleton):
    '''Singleton Class to mimic null'''


NO_THING = NoThing()
"""
Singleton Class instance that represents null object.
"""


def is_iterable(arg: object) -> bool:
    """
    Gets if ``arg`` is iterable.

    Args:
        arg (object): object to test

    Returns:
        bool: ``True`` if ``arg`` is an iterable object; Otherwise, ``False``.

    Note:
        if ``arg`` is of type str then return result is ``False``.

    Example:
        .. code-block:: python

            # non-string iterables    
            assert is_iterable(("f", "f"))    # tuple
            assert is_iterable(["f", "f"])    # list
            assert is_iterable(iter("ff"))    # iterator
            assert is_iterable(range(44))     # generator
            assert is_iterable(b"ff")         # bytes (Python 2 calls this a string)

            # strings or non-iterables
            assert not is_iterable(u"ff")     # string
            assert not is_iterable(44)        # integer
            assert not is_iterable(is_iterable)  # function
    """
    if isinstance(arg, str):
        return False
    result = False
    try:
        result = isinstance(iter(arg), Iterator)
    except Exception:
        result = False
    return result



# region Custom Errors
from inspect import isclass
from typing import Iterable, List, Type, Union
from ..helper import is_iterable
from ..rules import IRule

class CancelEventError(Exception):
    '''Cancel Event Error'''


class ReservedAttributeError(ValueError):
    '''Error when a reserved attribute is attempted to be set'''


class RuleError(Exception):
    '''Rule Error'''

    def __init__(self, **kwargs):
        """
        Constructor

        Keyword Arguments:
            err_rule (Type[IRule], optional): Rule that caused exception.
            rules_all (Iterable[Type[IRule]], optional): List of rules that were to all be matched.
                One of these rules is usually the reason this exception is being raised.
            rules_any (Iterable[Type[IRule]], optional): List of rules that required one or more matches.
                One of these rules is usually the reason this exception is being raised.
            arg_name (str, optional): Name of the argument for this exception.
            errors: (Union[Exception, Iterable[Exception]], optional): Exception or Exceptions
                that cause this error.
            fn_name: (str, optional): Name of function/property that raise error.
        """
        self._fn_name = kwargs.get('fn_name', None)
        self._err_rule = kwargs.get('err_rule', None)
        _rules = kwargs.get('rules_all', False)
        self._rules_all = self._get_rules(_rules)
        _rules = kwargs.get('rules_any', False)
        self._rules_any = self._get_rules(_rules)
        _rules = None
        self._arg_name = kwargs.get('arg_name', None)
        self._errors = kwargs.get('errors', None)

        msg = "RuleError:"
        if self._fn_name:
            msg = msg + f" '{self._fn_name}' error."
        if self._arg_name:
            msg = msg + f" Argument: '{self._arg_name}' failed validation."
        if self.err_rule and self._is_rule(self.err_rule):
            msg = msg + f"\nRule '{self.err_rule.__name__}' Failed validation."
        if len(self._rules_all) > 0:
            if len(self._rules_all) == 1:
                msg = msg + "\nExpected the following rule to match: "
            else:
                msg = msg + "\nExpected all of the following rules to match: "
            msg = msg + self._get_rules_str(self._rules_all) + "."
        if len(self._rules_any) > 0:
            if len(self._rules_any) == 1:
                msg = msg + "\nExpected the following rule to match: "
            else:
                msg = msg + "\nExpected at least one of the following rules to match: "
            msg = msg + self._get_rules_str(self._rules_any) + "."
        if self._is_errors() is True:
            msg = msg + "\nInner Error Message: " + self._get_inner_error_msg()
        self.message = msg
        super().__init__(self.message)

    # region private methods
    def _is_rule(self, rule) -> bool:
        if isclass(rule) and issubclass(rule, IRule):
            return True
        return False

    def _get_rules(self, rules: Iterable[IRule]) -> List[IRule]:
        result = []
        if rules and is_iterable(rules):
            for rule in rules:
                if self._is_rule(rule):
                    result.append(rule)
        return result

    def _get_rules_str(self, rules: List[IRule]) -> str:
        msg = ""
        for i, rule in enumerate(rules):
            if i > 0:
                msg = msg + ', '
            msg = f"{msg}{rule.__name__}"
        return msg

    def _is_errors(self) -> bool:
        if self._errors:
            if is_iterable(self._errors):
                return isinstance(self._errors[0], Exception)
            return isinstance(self._errors, Exception)
        return False

    def _get_first_error(self):
        if is_iterable(self._errors):
            return self._errors[0]
        return self._errors
    
    def _get_inner_error_msg(self) -> str:
        err = self._get_first_error()
        msg = err.__class__.__name__ + ": "
        msg = msg + str(err)
        return msg
    # endregion private methods

    # region Properties
    @property
    def arg_name(self) -> Union[str, None]:
        """Gets Name of the argument for this exception"""
        return self._arg_name

    @property
    def err_rule(self) -> Union[Type[IRule], None]:
        """Gets rule that caused exception."""
        return self._err_rule

    @property
    def errors(self) -> Union[Exception, Iterable[Exception], None]:
        """Gets Exception or Exceptions that cause this error"""
        return self._errors

    @property
    def rules_all(self) -> List[Type[IRule]]:
        """Gets list of rules that were to all be matched."""
        return self._rules_all

    @property
    def rules_any(self) -> List[Type[IRule]]:
        """Gets of rules that required one or more matches."""
        return self._rules_any
    
    @property
    def fn_name(self) -> Union[None, str]:
        """Gets the function/property name that raised the error"""
        return self._fn_name
    # endregion Properties
    # region Static Methods
    @staticmethod
    def from_rule_error(rule_error: 'RuleError', **kwargs) -> 'RuleError':
        """
        Creates a new RuleError from an existing RuleError
        
        Args:
            rule_error (RuleError): Current instance of RuleError use to base return value on.
        
        Keyword Arguments:
            kwargs: One or more Key, Value properties that will replace property of ``rule_error``.

        Returns:
            RuleError: New RuleError instance with updated properties included in ``**kwargs``.
        """
        rule_dict = {
            "err_rule": rule_error.err_rule,
            "rules_all": rule_error.rules_all,
            "rules_any": rule_error.rules_any,
            "arg_name": rule_error.arg_name,
            "errors": rule_error.errors,
            "fn_name": rule_error.fn_name
        }
        rule_dict.update({**kwargs})
        return RuleError(**rule_dict)
    # endregion Static Methods
# endregion Custom Errors



import enum
import functools
import inspect
import re
from typing import Dict, Iterable, Iterator, Optional, Set, Tuple, Union
from inspect import _ParameterKind, signature, isclass, Parameter, Signature
from ..checks import TypeChecker, RuleChecker
from ..rules import IRule
from ..helper import is_iterable
from ..exceptions import RuleError
from ..helper import NO_THING
from enum import IntEnum
# import wrapt


class DecFuncEnum(IntEnum):
    """Represents options for type of Function or Method"""
    FUNCTION = 1
    """Normal Unbound function"""
    METHOD_STATIC = 2
    """Class Static Method (@staticmethod)"""
    METHOD = 3
    """Class Method"""
    METHOD_CLASS = 4
    """Class Method (@classmethod)"""
    PROPERTY_CLASS = 5
    """Class Property (@property)"""

    def __str__(self):
        return self._name_


class _CommonBase(object):
    def __init__(self, **kwargs):
        """
        Constructor

        Keyword Arguments:
            opt_return (object, optional): Return value when decorator is invalid.
                By default an error is rasied when validation fails. If ``opt_return`` is
                supplied then it will be return when validation fails and no error will be raised.
        """
        self._opt_return = kwargs.get("opt_return", NO_THING)

    def _is_opt_return(self) -> bool:
        """
        Gets if opt_return value has been set in constructor

        Returns:
            bool: True if opt_return value is set; Otherwise, False
        """
        return not self._opt_return is NO_THING


class _DecBase(_CommonBase):
    _rx_star = re.compile("^\*(\d*)$")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._ftype: DecFuncEnum = kwargs.get("ftype", None)
        if self._ftype is not None:
            if not isinstance(self._ftype, DecFuncEnum):
                try:
                    self._ftype = DecFuncEnum(self._ftype)
                except:
                    raise TypeError(
                        f"{self.__class__.__name__} requires arg 'ftype' to be a 'DecFuncType")
        else:
            self._ftype = DecFuncEnum.FUNCTION
        self._cache = {}

    def _is_placeholder_arg(self, arg_name: str) -> bool:
        m = _DecBase._rx_star.match(arg_name)
        if m:
            return True
        return False

    def _drop_arg_first(self) -> bool:
        return self._ftype.value > DecFuncEnum.METHOD_STATIC.value

    def _get_args(self, args: Iterable[object]):
        if self._drop_arg_first():
            return args[1:]
        return args

    def _get_args_star(self, func: callable, args: Iterable[object]) -> Iterable[object]:
        """
        Get args accounting for ``*args`` postions in function and if function class method.

        Args:
            func (callable): function with args
            args (Iterable[object]): function current args

        Returns:
            Iterable[object]: New args that may be a subset of all of orignial ``args``.
        """
        pos = self._get_star_args_pos(func)
        drop_first = self._drop_arg_first()
        i = 0
        if pos > 0:
            i += pos
        if drop_first:
            i += 1
        if i > 0:
            return args[i:]
        return args

    def _get_signature(self, func) -> Signature:
        sig = self._cache.get("signature", False)
        if sig:
            return sig
        self._cache["signature"] = signature(func)
        return self._cache["signature"]

    def _get_args_dict(self, func: callable, args: Iterable[object], kwargs: Dict[str, object]) -> Dict[str, object]:
        # Positional argument cannot appear after keyword arguments
        # no *args after **kwargs def foo(**kwargs, *args) not valid
        # def foo(name, age, *args) not valid
        #
        # When only args are passed in (no **kwargs present) with named args then
        # the last named args with defaults are like additional args
        sig = self._get_signature(func)
        name_values = []
        i = 0
        args_pos = -1
        drop_first = self._drop_arg_first()
        for k, v in sig.parameters.items():
            if v.kind == _ParameterKind.VAR_POSITIONAL:  # args
                args_pos = i
                if drop_first:
                    args_pos -= 1
                continue
            if v.kind == _ParameterKind.VAR_KEYWORD:  # kwargs
                continue
            if not v.default is inspect._empty:
                name_values.append((k, v.default))
            else:
                name_values.append((k, NO_THING))
            i += 1
        arg_offset = 0 if args_pos == -1 else args_pos
        if drop_first and len(args) > arg_offset:
            _args = args[(arg_offset + 1):]
        else:
            _args = args[arg_offset:]
        arg_len = len(_args)
        name_values_len = len(name_values)
        if drop_first and name_values_len > 0:
            del name_values[0]
            name_values_len -= 1
        offset = 0
        if args_pos >= 0 and name_values_len > 0:
            # count how many name, values from default
            # are default values
            reversed_list = list(reversed(name_values))
            for k, v in reversed_list:
                if not v is NO_THING:
                    offset += 1
                else:
                    break

        name_defaults = {}
        # add first keys of name_values
        if args_pos > 0:
            # there are name values before *args in function
            # add first args so they are in the same order as entered.
            for j in range(args_pos):
                nv = name_values[j]
                name_defaults[nv[0]] = nv[1]
        # add args
        if args_pos >= 0:
            # add args that are positional arguments
            for j, arg in enumerate(_args):
                key = "*" + str(j + arg_offset)
                name_defaults[key] = arg

        if args_pos > 0:
            # add any remaining keys for name_values
            remaining = len(name_values) - args_pos
            if remaining > 0:
                for j in range(remaining):
                    index = j + args_pos
                    nv = name_values[index]
                    name_defaults[nv[0]] = nv[1]
        else:
            # there were no positional args before *args keyword so
            # all all the name, values after postitioan args.
            for k, v in name_values:
                name_defaults[k] = v
        if args_pos != 0 and arg_len > 0:
            # if *args are not in position 0 and there are *args then
            # make a dictionary of name values and update the
            # name_defaults dictionary.
            argnames = []
            for j in range(len(name_values) - offset):
                el = name_values[j]
                argnames.append(el[0])
            if drop_first:
                _zip_args = args[1:]
            else:
                _zip_args = args
            d = {**dict(zip(argnames, _zip_args[:len(argnames)]))}
            name_defaults.update(d)
        if len(kwargs) > 0:
            # update name_default with any kwargs that are passed in.
            # this will update any default values with the values passed
            # into function at call time.
            name_defaults.update(kwargs)
        return name_defaults

    def _get_formated_types(self, types: Iterator[type]) -> str:
        result = ''
        for i, t in enumerate(types):
            if i > 0:
                result = result + ' | '
            result = f"{result}{t}"
        return result

    def _get_ordinal(self, num: int) -> str:
        """
        Returns the ordinal number of a given integer, as a string.

        Args:
            num (int): integer to get ordinal value of.

        Returns:
            str: num as ordinal str. eg. 1 -> 1st, 2 -> 2nd, 3 -> 3rd, etc.
        """
        if 10 <= num % 100 < 20:
            return '{0}th'.format(num)
        else:
            ord = {1: 'st', 2: 'nd', 3: 'rd'}.get(num % 10, 'th')
            return '{0}{1}'.format(num, ord)

    def _get_star_args_pos(self, func: callable) -> int:
        """
        Gets the zero base postion of *args in a function.

        Args:
            func (callable): function to get *args postion of.

        Returns:
            int: -1 if  *args not present; Otherwise zero based postion of *args
        """
        result = self._cache.get("star_args_pos", None)
        if not result is None:
            return result
        sig = self._get_signature(func)
        args_pos = -1
        drop_first = self._drop_arg_first()
        i = 0
        for k, v in sig.parameters.items():
            if v.kind == _ParameterKind.VAR_POSITIONAL:  # args
                args_pos = i
                if drop_first:
                    args_pos -= 1
                break
            i += 1
        self._cache["star_args_pos"] = args_pos
        return self._cache["star_args_pos"]


class _RuleBase(_DecBase):
    def _get_err(self, fn: callable, e: RuleError):
        err = RuleError.from_rule_error(e, fn_name=fn.__name__)
        return err


class TypeCheck(_DecBase):
    """
    Decorator that decorates methods that requires args to match a type specificed in a list

    See Also:
        :doc:`../../usage/Decorator/TypeCheck`
    """

    def __init__(self, *args: Union[type, Iterable[type]], **kwargs):
        """
        Constructor

        Other Parameters:
            args (type): One or more types for wrapped function args to match.

        Keyword Arguments:
            raise_error: (bool, optional): If ``True`` then a ``TypeError`` will be raised if a
                validation fails. If ``False`` then an attribute will be set on decorated function
                named ``is_types_valid`` indicating if validation status.
                Default ``True``.
            type_instance_check (bool, optional): If ``True`` then args are tested also for ``isinstance()``
                if type does not match, rather then just type check. If ``False`` then values willl only be
                tested as type.
                Default ``True``
            ftype (DecFuncType, optional): Type of function that decorator is applied on.
                Default ``DecFuncType.FUNCTION``
            opt_return (object, optional): Return value when decorator is invalid.
                By default an error is rasied when validation fails. If ``opt_return`` is
                supplied then it will be return when validation fails and no error will be raised.

        Raises:
            TypeError: If ``types`` arg is not a iterable object such as a list or tuple.
            TypeError: If any arg is not of a type listed in ``types``.
        """
        super().__init__(**kwargs)
        self._tc = None
        self._types = [arg for arg in args]
        if kwargs:
            # keyword args are passed to TypeChecker
            self._kwargs = {**kwargs}
        else:
            self._kwargs = {}

    def __call__(self, func: callable):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            _args = self._get_args(args)
            try:
                is_valid = self._typechecker.validate(*_args, **kwargs)
                if self._typechecker.raise_error is False:
                    wrapper.is_types_valid = is_valid
                if is_valid is False and self._is_opt_return() is True:
                    return self._opt_return
            except TypeError as e:
                if self._is_opt_return():
                    return self._opt_return
                raise e
            return func(*args, **kwargs)
        if self._typechecker.raise_error is False:
            wrapper.is_types_valid = True
        return wrapper

    @property
    def _typechecker(self) -> TypeChecker:
        if self._tc is None:
            self._tc = TypeChecker(*self._types, **self._kwargs)
        return self._tc


class AcceptedTypes(_DecBase):
    """
    Decorator that decorates methods that requires args to match types specificed in a list

    See Also:
        :doc:`../../usage/Decorator/AcceptedTypes`
    """

    def __init__(self, *args: Union[type, Iterable[type]], **kwargs):
        """
        Constructor

        Other Parameters:
            args (Union[type, Iterable[type]]): One or more types or Iterator[type] for validation.

        Keyword Arguments:
            type_instance_check (bool, optional): If ``True`` then args are tested also for ``isinstance()``
                if type does not match, rather then just type check. If ``False`` then values willl only be
                tested as type.
                Default ``True``
            ftype (DecFuncType, optional): Type of function that decorator is applied on.
                Default ``DecFuncType.FUNCTION``
            opt_return (object, optional): Return value when decorator is invalid.
                By default an error is rasied when validation fails. If ``opt_return`` is
                supplied then it will be return when validation fails and no error will be raised.
        """
        super().__init__(**kwargs)
        self._tc = None
        self._types = []
        for arg in args:
            if is_iterable(arg):
                self._types.append(arg)
            else:
                self._types.append(tuple([arg]))
        if kwargs:
            # keyword args are passed to TypeChecker
            self._kwargs = {**kwargs}
        else:
            self._kwargs = {}

    def _get_formated_types(self, types: Iterator[type]) -> str:
        result = ''
        for i, t in enumerate(types):
            if i > 0:
                result = result + ' | '
            result = f"{result}{t}"
        return result

    def __call__(self, func: callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            arg_name_values = self._get_args_dict(func, args, kwargs)
            arg_keys = arg_name_values.keys()
            if len(arg_keys) is not len(self._types):
                if self._is_opt_return():
                    return self._opt_return
                raise ValueError(
                    'Invalid number of arguments for {0}()'.format(func.__name__))
            arg_type = zip(arg_keys, self._types)
            i = 0
            for arg_info in arg_type:
                key = arg_info[0]
                value = arg_name_values[key]
                tc = TypeChecker(*arg_info[1], **self._kwargs)
                # ensure errors are raised if not valid
                tc.raise_error = True
                if self._is_placeholder_arg(key):
                    try:
                        tc.validate(value)
                    except TypeError:
                        if self._is_opt_return():
                            return self._opt_return
                        raise TypeError(self._get_err_msg(name=None, value=value,
                                                          types=arg_info[1], arg_index=i,
                                                          fn=func))
                else:
                    try:
                        tc.validate(**{key: value})
                    except TypeError:
                        if self._is_opt_return():
                            return self._opt_return
                        raise TypeError(self._get_err_msg(name=key, value=value,
                                                          types=arg_info[1], arg_index=i,
                                                          fn=func))
                i += 1
            return func(*args, **kwargs)
        return wrapper

    def _get_err_msg(self, name: Union[str, None], value: object, types: Iterator[type], arg_index: int, fn: callable):
        str_types = self._get_formated_types(types=types)
        str_ord = self._get_ordinal(arg_index + 1)
        if self._ftype == DecFuncEnum.PROPERTY_CLASS:
            msg = f"'{fn.__name__}' property error. Arg '{name}' expected type of '{str_types}' but got '{type(value).__name__}'"
            return msg
        if name:
            msg = f"Arg '{name}' in {str_ord} position is expected to be of '{str_types}' but got '{type(value).__name__}'"
        else:
            msg = f"Arg in {str_ord} position of is expected to be of '{str_types}' but got '{type(value).__name__}'"
        return msg


class ArgsLen(_DecBase):
    """
    Decorartor that sets the number of args that can be added to a function

    Raises:
        ValueError: If wrong args are passed into construcor.
        ValueError: If validation of arg count fails.

    See Also:
        :doc:`../../usage/Decorator/ArgsLen`
    """

    def __init__(self, *args: Union[type, Iterable[type]], **kwargs):
        """
        Constructor

        Other Parameters:
            args (Union[int, iterable[int]]): One or more int or Iterator[int] for validation.

                * Single ``int`` values are to match exact.
                * ``iterable[int]`` must be a pair of ``int`` with the first ``int`` less then the second ``int``.

        Keyword Arguments:
            ftype (DecFuncType, optional): Type of function that decorator is applied on.
                Default ``DecFuncType.FUNCTION``
            opt_return (object, optional): Return value when decorator is invalid.
                By default an error is rasied when validation fails. If ``opt_return`` is
                supplied then it will be return when validation fails and no error will be raised.
        """
        super().__init__(**kwargs)
        self._ranges: Set[Tuple[int, int]] = set()
        self._lengths: Set[int] = set()
        for arg in args:
            if isinstance(arg, int):
                if arg >= 0:
                    self._lengths.add(arg)
            elif is_iterable(arg) and len(arg) == 2:
                arg1 = arg[0]
                arg2 = arg[1]
                if isinstance(arg1, int) and isinstance(arg2, int) \
                        and arg1 >= 0 and arg2 > arg1:
                    self._ranges.add((arg1, arg2))
        valid = len(self._lengths) > 0 or len(self._ranges) > 0
        if not valid:
            msg = f"{self.__class__.__name__} error. constructor must have valid args of of postive int and/or postive pairs of int."
            raise ValueError(msg)

    def _get_valid_counts(self) -> str:
        str_len = ""
        lengths = sorted(self._lengths)
        ranges = sorted(self._ranges)
        for i, length in enumerate(lengths):
            if i > 0:
                str_len = str_len + ", "
            str_len = str_len + str(length)
        str_rng = ""
        for i, rng in enumerate(ranges):
            if i > 0:
                str_rng = str_rng + ", "
            str_rng = str_rng + f"({rng[0]}, {rng[1]})"
        result = ""
        len_lengths = len(lengths)
        len_ranges = len(ranges)
        if len_lengths > 0:
            if len_lengths == 1:
                result = result + "Expected Length: "
            else:
                result = result + "Expected Lengths: "
            result = result + str_len + "."
        if len_ranges > 0:
            if len_lengths > 0:
                result = result + " "
            if len_ranges == 1:
                result = result + "Expected Range: "
            else:
                result = result + "Expected Ranges: "
            result = result + str_rng + "."
        return result

    def __call__(self, func: callable):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            _args = self._get_args_star(func=func, args=args)
            _args_len = len(_args)
            is_valid = False
            if _args_len >= 0:
                for i in self._lengths:
                    if _args_len == i:
                        is_valid = True
                        break
                if is_valid is False:
                    for range in self._ranges:
                        if _args_len >= range[0] and _args_len <= range[1]:
                            is_valid = True
                            break
            if is_valid is False:
                if self._is_opt_return():
                    return self._opt_return
                msg = f"Invalid number of args pass into '{func.__name__}'.\n{self._get_valid_counts()}"
                msg = msg + f" Got '{_args_len}' args."
                msg = msg + f"\n{self.__class__.__name__} decorator Error."
                raise ValueError(msg)
            return func(*args, **kwargs)
        # wrapper.is_types_valid = self.is_valid
        return wrapper


class ArgsMinMax(_DecBase):
    """
    Decorartor that sets the min and or max number of args that can be added to a function


    See Also:
        :doc:`../../usage/Decorator/ArgsMinMax`
    """

    def __init__(self, min: Optional[int]=0, max: Optional[int] = None, **kwargs):
        """
        Constructor

        Args:
            min (int, optional): Min number of args for a function. Defaults to 0.
            max (int, optional): Max number of args for a function. Defaults to None.

        Keyword Arguments:
            ftype (DecFuncType, optional): Type of function that decorator is applied on.
                Default ``DecFuncType.FUNCTION``
            opt_return (object, optional): Return value when decorator is invalid.
                By default an error is rasied when validation fails. If ``opt_return`` is
                supplied then it will be return when validation fails and no error will be raised.
        """
        super().__init__(**kwargs)
        self._min = int(min)
        if isinstance(max, int):
            self._max = max
        else:
            self._max = None
    
    def _get_min_max(self) -> Tuple[int, int]:
        _max = -1 if self._max is None else self._max
        _min = self._min
        return _min, _max
    
    def _get_valid_counts(self) -> str:

        _min, _max  = self._get_min_max()
        msg = ""
        if _min > 0:
            msg = msg + "Expected min of " + str(_min) + "."
        if _max >= 0:
            if _min > 0:
                msg = msg + " "
            msg = msg + "Expected max of " + str(_max) + "."
        return msg

    def _get_error_msg(self, func: callable, args_len: int) -> str:
        msg = f"Invalid number of args pass into '{func.__name__}'.\n{self._get_valid_counts()}"
        msg = msg + f" Got '{args_len}' args."
        msg = msg + f"\n{self.__class__.__name__} decorator Error."
        return msg
    
    def __call__(self, func: callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            _min, _max = self._get_min_max()
            has_rules = _min > 0 or _max >= 0
            if has_rules:
                _args = self._get_args_star(func=func, args=args)
                _args_len = len(_args)
                is_valid = True
                if _min > 0:
                    if _args_len < _min:
                        is_valid = False
                if is_valid == True and _max >= 0:
                    if _args_len > _max:
                        is_valid = False
                if is_valid is False:
                    if self._is_opt_return():
                        return self._opt_return
                    raise ValueError(self._get_error_msg(func=func,args_len=_args_len))
            return func(*args, **kwargs)
        # wrapper.is_types_valid = self.is_valid
        return wrapper

class ReturnRuleAll(_RuleBase):
    """
    Decorator that decorates methods that require return value to match all rules specificed.

    See Also:
        :doc:`../../usage/Decorator/ReturnRuleAll`
    """

    def __init__(self, *args: IRule, **kwargs):
        """
        Constructor

        Args:
            args (IRule): One or more rules to use for validation
        Keyword Arguments:
            ftype (DecFuncType, optional): Type of function that decorator is applied on.
                Default ``DecFuncType.FUNCTION``
                Default ``True``
            opt_return (object, optional): Return value when decorator is invalid.
                By default an error is rasied when validation fails. If ``opt_return`` is
                supplied then it will be return when validation fails and no error will be raised.
        """
        super().__init__(**kwargs)
        self._rc = None
        self._rules = [arg for arg in args]
        if kwargs:
            self._kwargs = {**kwargs}
        else:
            self._kwargs = {}

    def __call__(self, func: callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return_value = func(*args, **kwargs)
            rc = self._rulechecker
            try:
                rc.validate_all(**{"return": return_value})
            except RuleError as e:
                if self._is_opt_return():
                    return self._opt_return
                err = self._get_err(fn=func, e=e)
                raise err
            return return_value
        return wrapper

    @property
    def _rulechecker(self) -> RuleChecker:
        if self._rc is None:
            self._rc = RuleChecker(rules_all=self._rules, **self._kwargs)
            self._rc.raise_error = True
        return self._rc


class ReturnRuleAny(_RuleBase):
    """
    Decorator that decorates methods that require return value to match any of the rules specificed.

    See Also:
        :doc:`../../usage/Decorator/ReturnRuleAny`
    """

    def __init__(self, *args: IRule, **kwargs):
        """
        Constructor

        Args:
            args (IRule): One or more rules to use for validation
        Keyword Arguments:
            ftype (DecFuncType, optional): Type of function that decorator is applied on.
                Default ``DecFuncType.FUNCTION``
            opt_return (object, optional): Return value when decorator is invalid.
                By default an error is rasied when validation fails. If ``opt_return`` is
                supplied then it will be return when validation fails and no error will be raised.
        """
        super().__init__(**kwargs)
        self._rc = None
        self._rules = [arg for arg in args]
        if kwargs:
            self._kwargs = {**kwargs}
        else:
            self._kwargs = {}

    def __call__(self, func: callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return_value = func(*args, **kwargs)
            rc = self._rulechecker
            # rc.current_arg = "return"
            try:
                rc.validate_any(**{"return": return_value})
            except RuleError as e:
                if self._is_opt_return():
                    return self._opt_return
                err = self._get_err(fn=func, e=e)
                raise err
            return return_value
        return wrapper

    @property
    def _rulechecker(self) -> RuleChecker:
        if self._rc is None:
            self._rc = RuleChecker(rules_any=self._rules, **self._kwargs)
            self._rc.raise_error = True
        return self._rc


class ReturnType(_DecBase):
    """
    Decorator that decorates methods that require return value to match a type specificed.

    See Also:
        :doc:`../../usage/Decorator/ReturnType`
    """

    def __init__(self, *args: type, **kwargs):
        """
        Constructor

        Args:
            args (type): One ore more types that is used to validate return type.

        Keyword Arguments:
            type_instance_check (bool, optional): If ``True`` then args are tested also for ``isinstance()``
                if type does not match, rather then just type check. If ``False`` then values willl only be
                tested as type.
                Default ``True``
            opt_return (object, optional): Return value when decorator is invalid.
                By default an error is rasied when validation fails. If ``opt_return`` is
                supplied then it will be return when validation fails and no error will be raised.
        """
        super().__init__(**kwargs)
        self._tc = None
        self._types = [*args]
        if kwargs:
            # keyword args are passed to TypeChecker
            self._kwargs = {**kwargs}
        else:
            self._kwargs = {}

    def __call__(self, func: callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return_value = func(*args, **kwargs)
            try:
                self._typechecker.validate(return_value)
            except TypeError:
                if self._is_opt_return():
                    return self._opt_return
                # catch type error and raise a new one so a more fitting message is raised.
                raise TypeError(self._get_err_msg(return_value))
            return return_value
        return wrapper

    def _get_err_msg(self, value: object):
        str_types = self._get_formated_types(self._types)
        msg = f"Return Value is expected to be of '{str_types}' but got '{type(value).__name__}'"
        return msg

    @property
    def _typechecker(self) -> TypeChecker:
        if self._tc is None:
            self._tc = TypeChecker(*self._types, **self._kwargs)
            # ensure errors are raised if not valid
            self._tc.raise_error = True
        return self._tc


class TypeCheckKw(_DecBase):
    """
    Decorator that decorates methods that require key, value args to match a type specificed in a list

    See Also:
        :doc:`../../usage/Decorator/TypeCheckKw`
    """

    def __init__(self, arg_info: Dict[str, Union[int, type, Iterable[type]]], types: Optional[Iterable[Union[type, Iterable[type]]]] = None, **kwargs):
        """
        Constructor

        Args:
            arg_info (Dict[str, Union[int, type, Iterable[type]]]): Dictionary of Key and int, type, or Iterable[type].
                Each Key represents that name of an arg to match one or more types(s).
                If value is int then value is an index that corresponds to an item in ``types``.
            types (Iterable[Union[type, Iterable[type]]], optional): List of types for arg_info entries to match.
                Default ``None``

        Keyword Arguments:
            raise_error: (bool, optional): If ``True`` then a ``TypeError`` will be raised if a
                validation fails. If ``False`` then an attribute will be set on decorated function
                named ``is_types_kw_valid`` indicating if validation status.
                Default ``True``.
            type_instance_check (bool, optional): If ``True`` then args are tested also for ``isinstance()``
                if type does not match, rather then just type check. If ``False`` then values willl only be
                tested as type. Default ``True``
            ftype (DecFuncType, optional): Type of function that decorator is applied on.
                Default ``DecFuncType.FUNCTION``
            opt_return (object, optional): Return value when decorator is invalid.
                By default an error is rasied when validation fails. If ``opt_return`` is
                supplied then it will be return when validation fails and no error will be raised.
        """
        super().__init__(**kwargs)
        self._raise_error = bool(kwargs.get("raise_error", True))
        self._arg_index = arg_info
        if types is None:
            self._types = []
        else:
            self._types = types
        if kwargs:
            self._kwargs = {**kwargs}
        else:
            self._kwargs = {}

    def _get_types(self, key: str) -> Iterable:
        value = self._arg_index[key]
        if isinstance(value, int):
            t = self._types[value]
            if isinstance(t, Iterable):
                return t
            return [t]
        if is_iterable(value):
            return value
        else:
            # make iterable
            return (value,)

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            is_valid = True
            arg_name_values = self._get_args_dict(func, args, kwargs)
            arg_keys = arg_name_values.keys()
            tc = False
            for key in self._arg_index.keys():
                if key in arg_keys:
                    is_valid = False
                    types = self._get_types(key=key)
                    if len(types) == 0:
                        continue
                    value = arg_name_values[key]
                    tc = TypeChecker(*types, **self._kwargs)
                    try:
                        is_valid = tc.validate(**{key: value})
                        if is_valid is False:
                            break
                    except TypeError as e:
                        if self._is_opt_return():
                            return self._opt_return
                        raise e
            if tc and tc.raise_error is False:
                wrapper.is_types_kw_valid = is_valid
                if is_valid == False and self._is_opt_return() == True:
                    return self._opt_return
            return func(*args, **kwargs)
        if self._raise_error is False:
            wrapper.is_types_kw_valid = True
        return wrapper


class RuleCheckAny(_RuleBase):
    """
    Decorator that decorates methods that require args to match a rule specificed in ``rules`` list.

    If a function arg does not match at least one rule in ``rules`` list then validation will fail.

    See Also:
        :doc:`../../usage/Decorator/RuleCheckAny`
    """

    def __init__(self, *args: IRule, **kwargs):
        """
        Constructor

        Other Parameters:
            args (IRule): One or more rules to use for validation

        Keyword Arguments:
            raise_error (bool, optional): If ``True`` then an Exception will be raised if a
                validation fails. The kind of exception raised depends on the rule that is
                invalid. Typically a ``TypeError`` or a ``ValueError`` is raised.

                If ``False`` then an attribute will be set on decorated function
                named ``is_rules_any_valid`` indicating if validation status.
                Default ``True``.
            ftype (DecFuncType, optional): Type of function that decorator is applied on.
                Default ``DecFuncType.FUNCTION``
            opt_return (object, optional): Return value when decorator is invalid.
                By default an error is rasied when validation fails. If ``opt_return`` is
                supplied then it will be return when validation fails and no error will be raised.
        """
        super().__init__(**kwargs)
        self._raise_error = bool(kwargs.get("raise_error", True))
        self._rc = None
        self._rules = [arg for arg in args]
        if kwargs:
            self._kwargs = {**kwargs}
        else:
            self._kwargs = {}

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            _args = self._get_args(args)
            is_valid = False
            try:
                is_valid = self._rulechecker.validate_any(*_args, **kwargs)
            except RuleError as err:
                if self._is_opt_return():
                    return self._opt_return
                err_rule = self._get_err(fn=func, e=err)
                raise err_rule
            if self._raise_error is False:
                wrapper.is_rules_any_valid = is_valid
                if is_valid == False and self._is_opt_return() == True:
                    return self._opt_return
            return func(*args, **kwargs)
        if self._raise_error is False:
            wrapper.is_rules_any_valid = True
        return wrapper

    @property
    def _rulechecker(self) -> RuleChecker:
        if self._rc is None:
            self._rc = RuleChecker(rules_any=self._rules, **self._kwargs)
        return self._rc


class RuleCheckAll(_RuleBase):
    """
    Decorator that decorates methods that require args to match all rules specificed in ``rules`` list.

    If a function arg does not match all rules in ``rules`` list then validation will fail.

    See Also:
        :doc:`../../usage/Decorator/RuleCheckAll`
    """

    def __init__(self, *args: IRule, **kwargs):
        """
        Constructor

        Other Parameters:
            args (IRule): One or more rules to use for validation

        Keyword Arguments:
            raise_error (bool, optional): If ``True`` then an Exception will be raised if a
                validation fails. The kind of exception raised depends on the rule that is
                invalid. Typically a ``TypeError`` or a ``ValueError`` is raised.

                If ``False`` then an attribute will be set on decorated function
                named ``is_rules_all_valid`` indicating if validation status.
                Default ``True``.
            ftype (DecFuncType, optional): Type of function that decorator is applied on.
                Default ``DecFuncType.FUNCTION``
            opt_return (object, optional): Return value when decorator is invalid.
                By default an error is rasied when validation fails. If ``opt_return`` is
                supplied then it will be return when validation fails and no error will be raised.
        """
        super().__init__(**kwargs)
        self._raise_error = bool(kwargs.get("raise_error", True))
        self._rc = None
        self._rules = [arg for arg in args]
        if kwargs:
            self._kwargs = {**kwargs}
        else:
            self._kwargs = {}

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            _args = self._get_args(args)
            is_valid = False
            try:
                is_valid = self._rulechecker.validate_all(*_args, **kwargs)
            except RuleError as err:
                if self._is_opt_return():
                    return self._opt_return
                err_rule = self._get_err(fn=func, e=err)
                raise err_rule
            if self._rulechecker.raise_error is False:
                wrapper.is_rules_all_valid = is_valid
                if is_valid == False and self._is_opt_return() == True:
                    return self._opt_return
            return func(*args, **kwargs)
        if self._raise_error is False:
            wrapper.is_rules_all_valid = True
        return wrapper

    @property
    def _rulechecker(self) -> RuleChecker:
        if self._rc is None:
            self._rc = RuleChecker(rules_all=self._rules, **self._kwargs)
        return self._rc


class RuleCheckAllKw(_RuleBase):
    """
    Decorator that decorates methods that require specific args to match rules specificed in ``rules`` list.

    If a function specific args do not match all matching rules in ``rules`` list then validation will fail.

    See Also:
        :doc:`../../usage/Decorator/RuleCheckAllKw`
    """

    def __init__(self, arg_info: Dict[str, Union[int, IRule, Iterable[IRule]]], rules: Optional[Iterable[Union[IRule, Iterable[IRule]]]] = None, **kwargs):
        """
        Constructor

        Args:
            arg_info (Dict[str, Union[int, IRule, Iterable[IRule]]]): Dictionary of Key and int, IRule, or Iterable[IRule].
                Each Key represents that name of an arg to check with one or more rules.
                If value is int then value is an index that corresponds to an item in ``rules``.
            rules (Iterable[Union[IRule, Iterable[IRule]]], optional): List of rules for arg_info entries to match.
                Default ``None``

        Keyword Arguments:
            raise_error (bool, optional): If ``True`` then an Exception will be raised if a
                validation fails. The kind of exception raised depends on the rule that is
                invalid. Typically a ``TypeError`` or a ``ValueError`` is raised.

                If ``False`` then an attribute will be set on decorated function
                named ``is_rules_kw_all_valid`` indicating if validation status.
                Default ``True``.
            ftype (DecFuncType, optional): Type of function that decorator is applied on.
                Default ``DecFuncType.FUNCTION``
            opt_return (object, optional): Return value when decorator is invalid.
                By default an error is rasied when validation fails. If ``opt_return`` is
                supplied then it will be return when validation fails and no error will be raised.
        """
        super().__init__(**kwargs)
        self._raise_error = bool(kwargs.get("raise_error", True))
        self._arg_index = arg_info
        if rules is None:
            self._rules = []
        else:
            self._rules = rules
        if kwargs:
            self._kwargs = {**kwargs}
        else:
            self._kwargs = {}

    def _get_rules(self, key: str) -> Iterable:
        value = self._arg_index[key]
        if isinstance(value, int):
            r = self._rules[value]
            if isinstance(r, Iterable):
                return r
            return [r]
        if isclass(value) and issubclass(value, IRule):
            return (value,)
        return value

    def __call__(self, func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            is_valid = True
            arg_name_values = self._get_args_dict(func, args, kwargs)
            arg_keys = arg_name_values.keys()
            add_attrib = None
            for key in self._arg_index.keys():
                if key in arg_keys:
                    rules = self._get_rules(key=key)
                    if len(rules) == 0:
                        continue
                    value = arg_name_values[key]
                    rc = RuleChecker(rules_all=rules, **self._kwargs)
                    if add_attrib is None:
                        add_attrib = not rc.raise_error
                    is_valid = False
                    try:
                        is_valid = rc.validate_all(**{key: value})
                    except RuleError as err:
                        if self._is_opt_return():
                            return self._opt_return
                        err_rule = self._get_err(fn=func, e=err)
                        raise err_rule
                    if is_valid is False:
                        break
            if add_attrib:
                wrapper.is_rules_kw_all_valid = is_valid
                if is_valid == False and self._is_opt_return() == True:
                    return self._opt_return
            return func(*args, **kwargs)
        if self._raise_error is False:
            wrapper.is_rules_kw_all_valid = True
        return wrapper


class RuleCheckAnyKw(RuleCheckAllKw):
    """
    Decorator that decorates methods that require specific args to match rules specificed in ``rules`` list.

    If a function specific args do not match at least one matching rule in ``rules`` list then validation will fail.

    See Also:
        :doc:`../../usage/Decorator/RuleCheckAnyKw`
    """

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            is_valid = True
            arg_name_values = self._get_args_dict(func, args, kwargs)
            arg_keys = arg_name_values.keys()
            add_attrib = None
            for key in self._arg_index.keys():
                if key in arg_keys:
                    rules = self._get_rules(key=key)
                    if len(rules) == 0:
                        continue
                    value = arg_name_values[key]
                    rc = RuleChecker(rules_any=rules, **self._kwargs)
                    if add_attrib is None:
                        add_attrib = not self._raise_error
                    is_valid = False
                    try:
                        is_valid = rc.validate_any(**{key: value})
                    except RuleError as err:
                        if self._is_opt_return():
                            return self._opt_return
                        err_rule = self._get_err(fn=func, e=err)
                        raise err_rule
                    if is_valid is False:
                        break
            if add_attrib:
                wrapper.is_rules_any_valid = is_valid
                if is_valid == False and self._is_opt_return() == True:
                    return self._opt_return
            return func(*args, **kwargs)
        if self._raise_error is False:
            wrapper.is_rules_any_valid = True
        return wrapper


class RequireArgs(_DecBase):
    """
    Decorator that defines required args for ``**kwargs`` of a function.

    See Also:
        :doc:`../../usage/Decorator/RequireArgs`
    """

    def __init__(self, *args: str, **kwargs):
        """
        Constructor

        Other Parameters:
            args (type): One or more names of wrapped function args to require.

        Keyword Arguments:
            ftype (DecFuncType, optional): Type of function that decorator is applied on.
                Default ``DecFuncType.FUNCTION``
            opt_return (object, optional): Return value when decorator is invalid.
                By default an error is rasied when validation fails. If ``opt_return`` is
                supplied then it will be return when validation fails and no error will be raised.
        """
        super().__init__(**kwargs)
        self._args = []
        for arg in args:
            if isinstance(arg, str):
                self._args.append(arg)

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            arg_name_values = self._get_args_dict(func, args, kwargs)
            arg_keys = arg_name_values.keys()
            for key in self._args:
                if not key in arg_keys:
                    if self._is_opt_return():
                        return self._opt_return
                    raise ValueError(
                        f"'{func.__name__}', '{key}' is a required arg.")
            return func(*args, **kwargs)
        return wrapper


class DefaultArgs(_CommonBase):
    """
    Decorator that defines default values for ``**kwargs`` of a function.

    See Also:
        :doc:`../../usage/Decorator/DefaultArgs`
    """

    def __init__(self, **kwargs: Dict[str, object]):
        """
        Constructor

        Keyword Arguments:
            kwargs (Dict[str, object]): One or more Key, Value pairs to assign to wrapped function args as defaults.
        """
        super().__init__(**kwargs)
        self._kwargs = {**kwargs}

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            for key, value in self._kwargs.items():
                if not key in kwargs:
                    kwargs[key] = value
            return func(*args, **kwargs)
        return wrapper


def calltracker(func):
    """
    Decorator method that adds ``has_been_called`` attribute to decorated method.
    ``has_been_called`` is ``False`` if method has not been called.
    ``has_been_called`` is ``True`` if method has been called.

    Note:
        This decorator needs to be the topmost decorator applied to a method

    Example:
        .. code-block:: python

            >>> @calltracker
            >>> def foo(msg):
            >>>     print(msg)

            >>> print(foo.has_been_called)
            False
            >>> foo("Hello World")
            Hello World
            >>> print(foo.has_been_called)
            True

    See Also:
        :doc:`../../usage/Decorator/calltracker`
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        wrapper.has_been_called = True
        return func(*args, **kwargs)
    wrapper.has_been_called = False
    return wrapper


def callcounter(func):
    """
    Decorator method that adds ``call_count`` attribute to decorated method.
    ``call_count`` is ``0`` if method has not been called.
    ``call_count`` increases by 1 each time method is been called.

    Note:
        This decorator needs to be the topmost decorator applied to a method

    Example:
        .. code-block:: python

            >>> @callcounter
            >>> def foo(msg):
            >>>     print(msg)

            >>> print("Call Count:", foo.call_count)
            0
            >>> foo("Hello")
            Hello
            >>> print("Call Count:", foo.call_count)
            1
            >>> foo("World")
            World
            >>> print("Call Count:", foo.call_count)
            2

    See Also:
        :doc:`../../usage/Decorator/callcounter`
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        wrapper.call_count += 1
        return func(*args, **kwargs)
    wrapper.call_count = 0
    return wrapper


def singleton(orig_cls):
    """
    Decorator that makes a class a singleton class

    Example:
        .. code-block:: python

            @singleton
            class Logger:
                def log(self, msg):
                    print(msg)

            logger1 = Logger()
            logger2 = Logger()
            assert logger1 is logger

    See Also:
        :doc:`../../usage/Decorator/singleton`
    """
    orig_new = orig_cls.__new__
    instance = None

    @functools.wraps(orig_cls.__new__)
    def __new__(cls, *args, **kwargs):
        nonlocal instance
        if instance is None:
            instance = orig_new(cls, *args, **kwargs)
        return instance
    orig_cls.__new__ = __new__
    return orig_cls


class AutoFill:
    """
    Class decorator that replaces the ``__init__`` function with one that
    sets instance attributes with the specified argument names and
    default values. The original ``__init__`` is called with no arguments
    after the instance attributes have been assigned.

    Example:
        .. code-block:: python

            >>> @AutoFill('a', 'b', c=3)
            ... class Foo: pass
            >>> sorted(Foo(1, 2).__dict__.items())
            [('a', 1), ('b', 2), ('c', 3)]
    """
    # https://codereview.stackexchange.com/questions/142073/class-decorator-in-python-to-set-variables-for-the-constructor

    def __init__(self,  *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def __call__(self, cls):
        class Wrapped(cls):
            """Wrapped Class"""
        self._init(Wrapped)
        return Wrapped

    def _init(self, cls):
        argnames = self._args
        defaults = self._kwargs
        kind = Parameter.POSITIONAL_OR_KEYWORD
        signature = Signature(
            [Parameter(a, kind) for a in argnames]
            + [Parameter(k, kind, default=v) for k, v in defaults.items()])
        original_init = cls.__init__

        def init(self, *args, **kwargs):
            bound = signature.bind(*args, **kwargs)
            bound.apply_defaults()
            for k, v in bound.arguments.items():
                setattr(self, k, v)
            original_init(self)

        cls.__init__ = init


class AutoFillKw:
    """
    Class decorator that replaces the ``__init__`` function with one that
    sets instance attributes with the specified key, value of ``kwargs``.
    The original ``__init__`` is called with any ``*args``
    after the instance attributes have been assigned.

    Example:
        .. code-block:: python

            >>> @AutoFillKw
            ... class Foo: pass
            >>> sorted(Foo(a=1, b=2, End="!").__dict__.items())
            [('End', '!'), ('a', 1), ('b', 2)]
    """

    def __init__(self, cls):
        self._cls = cls

    def __call__(self, *args, **kwargs):
        kind = Parameter.KEYWORD_ONLY
        signature = Signature(
            [Parameter(k, kind, default=v) for k, v in kwargs.items()])
        original_init = self._cls.__init__

        def init(self, *arguments, **kw):
            bound = signature.bind(**kw)
            bound.apply_defaults()
            for k, v in bound.arguments.items():
                setattr(self, k, v)
            original_init(self, *arguments)

        self._cls.__init__ = init
        return self._cls(*args, **kwargs)



# coding: utf-8
from abc import ABC, abstractmethod
import numbers
from typing import Optional
from ..helper import is_iterable
# region Interface


class IRule(ABC):
    """
    Abstract Interface Class for rules

    See Also:
        :doc:`/source/general/rules`
    """

    def __init__(self, key: str, name: str, value: object, raise_errors: bool, originator: object):
        """
        Constructor

        Args:
            key (str): the key that rule is to apply to.
            name (str): the name of the field that value was assigned
            value (object): the value that is assigned to ``field_name``
            raise_errors (bool): determinins if rule could raise an error when validation fails
            originator (object): the object that attributes validated for

        Raises:
            TypeError: If any arg is not of the correct type
        """
        if not isinstance(name, str):
            msg = self._get_type_error_msg(name, 'name', 'str')
            raise TypeError(msg)
        self._name: str = name

        if not isinstance(key, str):
            msg = self._get_type_error_msg(key, 'key', 'str')
            raise TypeError(msg)
        self._key: str = key

        if not isinstance(raise_errors, bool):
            msg = self._get_type_error_msg(
                raise_errors, 'raise_errors', 'bool')
            raise TypeError(msg)
        self._raise_errors = raise_errors

        self._value: object = value
        self._originator: object = originator

    # region Abstract Methods
    @abstractmethod
    def validate(self) -> bool:
        '''Gets attrib field and value are valid'''
    # endregion Abstract Methods

    def _get_type_error_msg(self, arg: Optional[object] = None, arg_name: Optional[str] = None, expected_type: Optional[str] = None) -> str:
        _arg = self.field_value if arg is None else arg
        _arg_name = self.key if arg_name is None else arg_name
        if expected_type:
            msg = f"Argument Error: '{_arg_name}' is expecting type of '{expected_type}'. Got type of '{type(_arg).__name__}'"
        else:
            msg = f"Argument Error: '{_arg_name}' is not expecting '{type(_arg).__name__}'"
        return msg

    def _get_not_type_error_msg(self, arg: Optional[object] = None, arg_name: Optional[str] = None, not_type: Optional[str] = None) -> str:
        _arg = self.field_value if arg is None else arg
        _arg_name = self.key if arg_name is None else arg_name
        if not_type:
            msg = f"Argument Error: '{_arg_name}' is expecting non '{not_type}'. Got type of '{type(_arg).__name__}'"
        else:
            msg = f"Argument Error: '{_arg_name}' is expecting non '{type(_arg).__name__}'."
        return msg
    # region Properties

    @property
    def field_name(self) -> str:
        '''
        Name of the field assigned.

        :getter: Gets the name of the field assigned
        :setter: Sets the name of the field assigned
        '''
        return self._name

    @field_name.setter
    def field_name(self, value: str):
        if not isinstance(value, str):
            msg = self._get_type_error_msg(value, 'field_name', 'str')
            raise TypeError(msg)
        self._name = value

    @property
    def field_value(self) -> object:
        """
        The value assigned to ``field_name``

        :getter: Gets value assigned to ``field_name``
        :setter: Sets value assigned to ``field_name``
        """
        return self._value

    @field_value.setter
    def field_value(self, value: object):
        self._value = value

    @property
    def key(self) -> str:
        '''Gets the key currently being read'''
        return self._key

    @key.setter
    def key(self, value: str):
        if not isinstance(value, str):
            msg = self._get_type_error_msg(value, 'key', 'str')
            raise TypeError(msg)
        self._key = value

    @property
    def raise_errors(self) -> bool:
        """
        Determines if a rule can raise an error when validation fails.

        :getter: Gets if a rule could raise an error when validation fails
        :setter: Sets if a rule could raise an error when validation fails
        """
        return self._raise_errors

    @raise_errors.setter
    def raise_errors(self, value: bool):
        if not isinstance(value, bool):
            msg = self._get_type_error_msg(value, 'raise_errors', 'bool')
            raise TypeError(msg)
        self._raise_errors = value

    @property
    def originator(self) -> object:
        '''Gets object that attributes validated for'''
        return self._originator
    # endregion Properties
# endregion Interface

# region Attrib rules


class RuleAttrNotExist(IRule):
    '''
    Rule to ensure an attribute does not exist before it is added to class.
    '''

    def validate(self) -> bool:
        """
        Validates that ``field_name`` is not an existing attribute of ``originator`` instance.

        Raises:
            AttributeError: If ``raise_errors`` is ``True`` and ``field_name`` is already an attribue of ``originator`` instance.

        Returns:
            bool: ``True`` if ``field_name`` is not an existing attribue of ``originator`` instance;
            Otherwise, ``False``.
        """
        result = not hasattr(self.originator, self.field_name)
        if result == False and self.raise_errors == True:
            raise AttributeError(
                f"'{self.field_name}' attribute already exist in current instance of '{type(self.originator).__name__}'")
        return result


class RuleAttrExist(IRule):
    '''
    Rule to ensure an attribute does exist before its value is set.
    '''

    def validate(self) -> bool:
        """
        Validates that ``field_name`` is an existing attribute of ``originator`` instance.

        Raises:
            AttributeError: If ``raise_errors`` is ``True`` and ``field_name`` is not an attribue of ``originator`` instance.

        Returns:
            bool: ``True`` if ``field_name`` is an existing attribue of ``originator`` instance;
            Otherwise, ``False``.
        """
        result = hasattr(self.originator, self.field_name)
        if result == False and self.raise_errors == True:
            raise AttributeError(
                f"'{self.field_name}' attribute does not exist in current instance of '{type(self.originator).__name__}'")
        return result
# endregion Attrib rules

# region None


class RuleNone(IRule):
    '''
    Rule that matched only if value is ``None``.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign to attribute is ``None``.

        Raises:
            ValueError: If ``raise_errors`` is ``True`` and ``field_value`` is not ``None``.

        Returns:
            bool: ``True`` if ``field_value`` is ``None``; Otherwise, ``False``.
        """
        if self.field_value is not None:
            if self.raise_errors:
                raise ValueError(
                    f"Arg error: {self.key} must be assigned a value")
            return False
        return True

class RuleNotNone(IRule):
    '''
    Rule that matched only if value is not ``None``.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign to attribute is not ``None``.

        Raises:
            ValueError: If ``raise_errors`` is ``True`` and ``field_value`` is ``None``.

        Returns:
            bool: ``True`` if ``field_value`` is not ``None``; Otherwise, ``False``.
        """
        if self.field_value is None:
            if self.raise_errors:
                raise ValueError(
                    f"Arg error: {self.key} must be assigned a value")
            return False
        return True

# endregion None

# region Number


class RuleNumber(IRule):
    '''
    Rule that matched only if value is a valid number.

    Note:
        If value is a of type ``bool`` then validation will fail for this rule.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign is a number

        Raises:
            TypeError: If ``raise_errors`` is ``True`` and ``field_value`` is not a number.

        Returns:
            bool: ``True`` if ``field_value`` is a number; Otherwise, ``False``.
        """
        # isinstance(False, int) is True
        # print(int(True)) 1
        # print(int(False)) 0
        if not isinstance(self.field_value, numbers.Number) or isinstance(self.field_value, bool):
            if self.raise_errors:
                raise TypeError(self._get_type_error_msg(
                    self.field_value, self.key, 'Number'))
            return False
        return True

# region Integer


class RuleInt(IRule):
    '''
    Rule that matched only if value is instance of ``int``.

    Note:
        If value is a of type ``bool`` then validation will fail for this rule.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign is an int

        Raises:
            TypeError: If ``raise_errors`` is ``True`` and ``field_value`` is not an int.

        Returns:
            bool: ``True`` if ``field_value`` is an ``int``; Otherwise, ``False``.
        """
        # isinstance(False, int) is True
        # print(int(True)) 1
        # print(int(False)) 0
        if not isinstance(self.field_value, int) or isinstance(self.field_value, bool):
            if self.raise_errors:
                raise TypeError(self._get_type_error_msg(expected_type='int'))
            return False
        return True


class RuleIntZero(RuleInt):
    '''
    Rule that matched only if value is equal to ``0``.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign is equal to ``0`` int.

        Raises:
            ValueError: If ``raise_errors`` is ``True`` and ``field_value`` is not equal to ``0`` int.

        Returns:
            bool: ``True`` if ``field_value`` equals ``0`` int; Otherwise, ``False``.
        """
        if not super().validate():
            return False
        if self.field_value != 0:
            if self.raise_errors:
                raise ValueError(
                    f"Arg error: '{self.key}' must be equal to 0 int value")
            return False
        return True

class RuleIntPositive(RuleInt):
    '''
    Rule that matched only if value is equal or greater than ``0``.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign is a posivite int

        Raises:
            TypeError: If ``raise_errors`` is ``True`` and ``field_value`` is not a positive int.

        Returns:
            bool: ``True`` if ``field_value`` is a positive int; Otherwise, ``False``.
        """
        if not super().validate():
            return False
        if self.field_value < 0:
            if self.raise_errors:
                raise ValueError(
                    f"Arg error: '{self.key}' must be a positive int value")
            return False
        return True


class RuleIntNegative(RuleInt):
    '''
    Rule that matched only if value is less than ``0``.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign is a negative int

        Raises:
            TypeError: If ``raise_errors`` is ``True`` and ``field_value`` is not a negative int.

        Returns:
            bool: ``True`` if ``field_value`` is a negative int; Otherwise, ``False``.
        """
        if not super().validate():
            return False
        if self.field_value >= 0:
            if self.raise_errors:
                raise ValueError(
                    f"Arg error: '{self.key}' must be a negative int value")
            return False
        return True


class RuleIntNegativeOrZero(RuleInt):
    '''
    Rule that matched only if value is equal or less than ``0``.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign is equal to zero or a negative int

        Raises:
            TypeError: If ``raise_errors`` is ``True`` and ``field_value`` is not a negative int.

        Returns:
            bool: ``True`` if ``field_value`` is equal to zero or a negative int; Otherwise, ``False``.
        """
        if not super().validate():
            return False
        if self.field_value > 0:
            if self.raise_errors:
                raise ValueError(
                    f"Arg error: '{self.key}' must be equal to zero or a negative int value")
            return False
        return True


class RuleByteUnsigned(RuleInt):
    '''
    Unsigned Byte rule, range from ``0`` to ``255``.
    '''

    def validate(self) -> bool:
        """
        Valids

        Raises:
            ValueError: If ``raise_errors`` is ``False`` and value is less then ``0`` or greater than ``255``.

        Returns:
            bool: ``True`` if Validation passes; Otherwise, ``False``.
        """
        if not super().validate():
            return False
        if self.field_value < 0 or self.field_value > 255:
            if self.raise_errors:
                raise ValueError(
                    f"Arg error: '{self.key}' must be a num from 0 to 255")
            return False
        return True


class RuleByteSigned(RuleInt):
    '''
    Signed Byte rule, range from ``-128`` to ``127``.
    '''

    def validate(self) -> bool:
        """
        Valids

        Raises:
            ValueError: If ``raise_errors`` is ``False`` and value is less then ``-128`` or greater than ``128``.

        Returns:
            bool: ``True`` if Validation passes; Otherwise, ``False``.
        """
        if not super().validate():
            return False
        if self.field_value < -128 or self.field_value > 127:
            if self.raise_errors:
                raise ValueError(
                    f"Arg error: '{self.key}' must be a num from -128 to 127")
            return False
        return True
# endregion Integer

# region Float Rules


class RuleFloat(IRule):
    '''
    Rule that matched only if value is to type ``float``.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign is a float

        Raises:
            TypeError: If ``raise_errors`` is ``True`` and ``field_value`` is not a float.

        Returns:
            bool: ``True`` if ``field_value`` is a positive float; Otherwise, ``False``.
        """
        if not isinstance(self.field_value, float):
            if self.raise_errors:
                raise TypeError(self._get_type_error_msg(
                    self.field_value, self.key, 'float'))
            return False
        return True


class RuleFloatZero(RuleFloat):
    '''
    Rule that matched only if value is equal to ``0.0``.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign equals ``0.0`` float

        Raises:
            ValueError: If ``raise_errors`` is ``True`` and ``field_value`` is not equal to ``0.0`` float.

        Returns:
            bool: ``True`` if ``field_value`` equals ``0.0`` float; Otherwise, ``False``.
        """
        if not super().validate():
            return False
        if self.field_value != 0.0:
            if self.raise_errors:
                raise ValueError(
                    f"Arg error: '{self.key}' must be equal to 0.0 float value")
            return False
        return True

class RuleFloatPositive(RuleFloat):
    '''
    Rule that matched only if value is equal or greater than ``0.0``.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign is a positive float

        Raises:
            ValueError: If ``raise_errors`` is ``True`` and ``field_value`` is not a positive float.

        Returns:
            bool: ``True`` if ``field_value`` is a positive float; Otherwise, ``False``.
        """
        if not super().validate():
            return False
        if self.field_value < 0.0:
            if self.raise_errors:
                raise ValueError(
                    f"Arg error: '{self.key}' must be a positive float value")
            return False
        return True


class RuleFloatNegative(RuleFloat):
    '''
    Rule that matched only if value is less than ``0.0``.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign is a negative float

        Raises:
            ValueError: If ``raise_errors`` is ``True`` and ``field_value`` is not a negative float.

        Returns:
            bool: ``True`` if ``field_value`` is a negative float; Otherwise, ``False``.
        """
        if not super().validate():
            return False
        if self.field_value >= 0.0:
            if self.raise_errors:
                raise ValueError(
                    f"Arg error: '{self.key}' must be a negative float value")
            return False
        return True


class RuleFloatNegativeOrZero(RuleFloat):
    '''
    Rule that matched only if value is equal or less than ``0.0``.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign is equal to ``0.0`` or a negative float

        Raises:
            ValueError: If ``raise_errors`` is ``True`` and ``field_value`` is not a negative float.

        Returns:
            bool: ``True`` if ``field_value`` is equal to ``0.0`` or a negative float; Otherwise, ``False``.
        """
        if not super().validate():
            return False
        if self.field_value > 0.0:
            if self.raise_errors:
                raise ValueError(
                    f"Arg error: '{self.key}' must be equal to 0.0 or a negative float value")
            return False
        return True
# endregion Float Rules

# endregion Number

# region String


class RuleStr(IRule):
    '''
    Rule that matched only if value is of type ``str``.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign is a string

        Raises:
            TypeError: If ``raise_errors`` is ``True`` and ``field_value`` is not instance of string.

        Returns:
            bool: ``True`` if ``field_value`` is a string; Otherwise, ``False``.
        """
        if not isinstance(self.field_value, str):
            if self.raise_errors:
                raise TypeError(self._get_type_error_msg(
                    self.field_value, self.key, 'str'))
            return False
        return True


class RuleStrEmpty(RuleStr):
    '''
    Rule that matched only if value is equal to empty string.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign is a string and is an empty string.

        Raises:
            ValueError: If ``raise_errors`` is ``True`` and ``field_value``
                is not an empty string.

        Returns:
            bool: ``True`` if value is an empty string; Otherwise; ``False``.
        """
        if not super().validate():
            return False
        value = self.field_value
        if len(value) != 0:
            if self.raise_errors:
                raise ValueError(
                    f"Arg error: {self.key} must be empty str")
            return False
        return True


class RuleStrNotNullOrEmpty(RuleStr):
    '''
    Rule that matched only if value is not ``None`` or empty string.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign is a string and is not a empty string.

        Raises:
            ValueError: If ``raise_errors`` is ``True`` and ``field_value``
                is not instance of string or is empty string

        Returns:
            bool: ``True`` if value is valid; Otherwise; ``False``.
        """
        if not super().validate():
            return False
        value = self.field_value
        if len(value) == 0:
            if self.raise_errors:
                raise ValueError(
                    f"Arg error: {self.key} must not be empty str")
            return False
        return True


class RuleStrNotNullEmptyWs(RuleStrNotNullOrEmpty):
    '''
    Rule that matched only if value is not ``None``, empty or whitespace.
    '''

    def validate(self) -> bool:
        """
        Validates that value to assign is a string and is not a empty or whitespace string.

        Raises:
            ValueError: If ``raise_errors`` is ``True`` and ``field_value``
                is not instance of string or is empty or whitespace string

        Returns:
            bool: ``True`` if value is valid; Otherwise; ``False``.
        """
        if not super().validate():
            return False
        value = self.field_value.strip()
        if len(value) == 0:
            if self.raise_errors:
                raise ValueError(
                    f"Arg error: '{self.key}' must not be empty or whitespace str")
            return False
        return True
# endregion String

# region boolean


class RuleBool(IRule):
    """
     Rule that matched only if value is instance of bool.
    """

    def validate(self) -> bool:
        """
        Validates that value to assign is a bool

        Raises:
            TypeError: If ``raise_errors`` is ``True`` and ``field_value`` is not instance of bool.

        Returns:
            bool: ``True`` if ``field_value`` is a bool; Otherwise, ``False``.
        """
        if not isinstance(self.field_value, bool):
            if self.raise_errors:
                raise TypeError(self._get_type_error_msg(expected_type='bool'))
            return False
        return True
# endregion boolean

# region Iterable


class RuleIterable(IRule):
    """
     Rule that matched only if value is iterable such as list, tuple, set.
    """

    def validate(self) -> bool:
        """
        Validates that value to assign is iterable

        Raises:
            TypeError: If ``raise_errors`` is ``True`` and ``field_value`` is not iterable.

        Returns:
            bool: ``True`` if ``field_value`` is a iterable; Otherwise, ``False``.
        """
        if not is_iterable(self.field_value):
            if self.raise_errors:
                raise TypeError(self._get_type_error_msg(expected_type="iterable"))
            return False
        return True


class RuleNotIterable(IRule):
    """
     Rule that matched only if value is not iterable.
    """

    def validate(self) -> bool:
        """
        Validates that value to assign is not iterable

        Raises:
            TypeError: If ``raise_errors`` is ``True`` and ``field_value`` is iterable.

        Returns:
            bool: ``True`` if ``field_value`` is a not iterable; Otherwise, ``False``.
        """
        if is_iterable(self.field_value):
            if self.raise_errors:
                raise TypeError(self._get_not_type_error_msg(not_type="iterable"))
            return False
        return True
    
    
# endregion Iterable



#!/usr/bin/env python
import pathlib
import os
from setuptools import setup
from kwhelp import __version__
PKG_NAME = 'kwargshelper'
VERSION = __version__

# The directory containing this file
HERE = pathlib.Path(__file__).parent
# The text of the README file
with open(HERE / "README.rst") as fh:
    README = fh.read()


def getListOfFiles(dir_name):
    '''
    For the given path, get the List of all files in the directory tree 
    '''
    # create a list of file and sub directories
    # names in the given directory
    list_of_files = os.listdir(dir_name)
    all_files = list()
    # Iterate over all the entries
    for entry in list_of_files:
        # Create full path
        full_path = os.path.join(dir_name, entry)
        # If entry is a directory then get the list of files in this directory
        if os.path.isdir(full_path):
            all_files = all_files + getListOfFiles(full_path)
        else:
            all_files.append(full_path)

    return all_files


def get_src_modules(root_path: pathlib.Path):
    dir_name = root_path / 'kwhelp'
    # Get the list of all files in directory tree at given path
    listOfFiles = getListOfFiles(dir_name)
    _slice = len(str(dir_name)) + 1
    py_lst = [f[_slice:-3]
              for f in list(filter(lambda p: p.endswith('.py'), listOfFiles))]
    return py_lst


MODULES = get_src_modules(HERE)


# This call to setup() does all the work
setup(
    name=PKG_NAME,
    version=VERSION,
    python_requires='>=3.6.0',
    description="Manages testing for valid kwargs key, values pairs and assigns class attributes for pairs.",
    long_description_content_type="text/x-rst",  # "text/markdown",
    long_description=README,
    url="https://github.com/Amourspirit/python-kwargshelper",
    author=":Barry-Thomas-Paul: Moss",
    author_email='bigbytetech@gmail.com',
    license="MIT",
    # package_dir={'kwhelp': 'kwhelp'},
    packages=['kwhelp', 'kwhelp.rules', 'kwhelp.helper',
              'kwhelp.decorator', 'kwhelp.exceptions', 'kwhelp.checks'],
    py_modules=MODULES,
    keywords=['python', 'kwargs', 'args', 'parse', 'helper'],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    include_package_data=True,
)
